
/* Modified for multicellcular version 
   using struct
   can be called from an iterpreter languages
   use cuda approach to accelerate calculation
   GNU public license
   Copyright to Bum-Rak Choi
   Oct 3, 2012 */

 
#ifndef LuoRudy1_H
#define LuoRudy1_H

#include "StcCell.h"

#ifdef __cplusplus 
	extern "C" {
#endif

/* List of variables and paramaters */ 
#define PI 3.1415926535897932846264



		
typedef double TypeCellML;


/* The Luo-Rudy Dynamic (LRd) Model of the Mammalian Ventricular Myocyte */
/* Gregory Faber */
/* Detailed list of equations and model description are provided in */

/*  Circ Res 1991;68:1501-1526      */ 
/*  Circ Res 1994;74:1071-1096      */


typedef struct {
	/* Terms for Solution of Conductance and Reversal Potential */
	TypeCellML R, frdy, temp, Rtemp, ORtemp, RtempOfrdy, frdyORtemp, frdyfrdyORtemp;
} LuoRudy1CM_Const ;


typedef struct  {
//  cell Constants
	
	TypeCellML CM_V; // 0	
	TypeCellML CM_xsize, CM_ysize, CM_zsize, CM_C, CM_Area ; // 1
	TypeCellML CM_Cm, CM_CmJ, CM_StimPeriod, CM_StimDuration, CM_StimCurrent, CM_StimOffset ; // 6
	
//	starting [12]	
	TypeCellML CM_gNa, CM_gammaNai, CM_gammaNao, CM_gammaKi, CM_gammaKo ;
	TypeCellML CM_PCa, CM_PNa, CM_PK ;
	TypeCellML CM_gammaCai, CM_gammaCao ;
	// starting [22]
	TypeCellML CM_Km_Ca, CM_gK_max, CM_PR_NaK, CM_gK1_max, CM_gKp, CM_K_mpCa, CM_I_pCa, CM_gNab ;
	// [30]
	TypeCellML CM_gCab,CM_I_NaK, CM_K_mNai, CM_K_mKo, CM_P_ns_Ca, CM_K_m_ns_Ca, CM_K_NaCa, CM_K_mNa ; 
	// [39]
	TypeCellML CM_K_mCa, CM_K_sat, CM_eta, CM_G_relMax, CM_tau_on, CM_tau_off, CM_t_CICR, CM_tau_tr ; 
	// [47]
	TypeCellML CM_K_mrel, CM_K_mup, CM_I_up, CM_Ca_NSR_max, CM_DeltaCa_i2, CM_DeltaCa_ith, CM_Am ;
	// [54]
	TypeCellML CM_V_myo, CM_V_JSR, CM_V_NSR, CM_gK, CM_gK1, CM_sigma, CM_G_relPeak, CM_K_leak , CM_G_rel ;

//  state Value 
	// [63]
	TypeCellML CM_Nai, CM_xm, CM_xh, CM_xj ;
	TypeCellML CM_Cai, CM_Ki, CM_xd, CM_xf, CM_xx;
	TypeCellML CM_Ca_JSR, CM_Ca_NSR ;
	// [74]
	TypeCellML dCM_V, dCM_Nai, dCM_xm, dCM_xh, dCM_xj ;
	TypeCellML dCM_Cai, dCM_Ki, dCM_xd, dCM_xf, dCM_xx;
	TypeCellML dCM_Ca_JSR, dCM_Ca_NSR ;
	
// ALGEBRIC 
	// [86]
    TypeCellML CM_dV_dt, CM_dVf_dt, CM_i_Na, CM_i_Ca_L, CM_i_K, CM_i_NaCa, CM_i_K1, CM_i_Kp ; 
    // [94]
    TypeCellML CM_i_p_Ca, CM_i_Na_b, CM_i_Ca_b, CM_xi_NaK, CM_i_ns_Ca; 
    //[99]
    TypeCellML CM_E_Na, CM_alpha_m, CM_beta_m, CM_alpha_h, CM_beta_h, CM_alpha_j, CM_beta_j, CM_xi_CaCa, CM_xi_CaK;
    //[108]
    TypeCellML CM_xi_CaNa, CM_I_CaCa, CM_I_CaK, CM_I_CaNa ; 
    // [112]
    TypeCellML CM_f_Ca, CM_alpha_d, CM_beta_d, CM_d_infinity, CM_tau_d, CM_alpha_f, CM_beta_f, CM_f_infinity, CM_tau_f, CM_E_K ;
    //[122]
    TypeCellML CM_Xi, CM_alpha_X, CM_beta_X, CM_E_K1, CM_K1_infinity, CM_alpha_K1, CM_beta_K1, CM_E_Kp, CM_Kp ; 
    //[131]
    TypeCellML CM_E_NaN, CM_E_CaN, CM_f_NaK, CM_xi_ns_Na, CM_xi_ns_K, CM_I_ns_Na, CM_I_ns_K; 
    //[138]
    TypeCellML CM_Vns, CM_EnsCa, CM_i_rel, CM_xi_up, CM_i_leak, CM_i_tr ;   

//CM Parameter Modification factors
	//[144]
	TypeCellML CM_IK1_value, CM_ICaL_value, CM_INa_value, CM_IK_value, CM_IKp_value ;

	// Ko Nao;
	TypeCellML Ko, Nao, Cao;

	 TypeCellML CM_I_st, CM_I_diff;
} LuoRudy1CM_State;


StcCell *prototype_Cell_LuoRudy1CM(void);
int alloc_Cell_LuoRudy1CM_device(StcCell *pStcCell);
void init_Cell_LuoRudy1CM_host(TypeCellML *pState, TypeCellML *pConst, int num);
void init_Cell_LuoRudy1CM_device(StcCell *pStcCell);
void iter_Cell_LuoRudy1CM(StcCell *pStcCell, TypeCellML deltaT);

#ifdef __cplusplus
}
#endif

#endif


