/* functions */

#include <math.h>
#include <cuda.h>
#include "Cell_LuoRudy1CM.h"

#define testZero(X) ( ( (X)>-1.0e-5 ) && ( (X) < 1.0e-5 ) )



StcCell *prototype_Cell_LuoRudy1CM(void) // StcCell *pStcCell)
{
	StcCell *pStcCell;
	pStcCell = (StcCell *)malloc(sizeof(StcCell));
	pStcCell->N_ALGEBRAIC = 0; // not using
	pStcCell->N_RATE = 0; // not using
	pStcCell->N_STATE_ORG = sizeof(LuoRudy1CM_State)/sizeof(TypeCellML) - 2; 
	pStcCell->N_STATE = pStcCell->N_STATE_ORG + 2; /* additional two for diffusion and stimulation */
	pStcCell->N_CONST = sizeof(LuoRudy1CM_Const)/sizeof(TypeCellML);
	pStcCell->LOC_STIM = pStcCell->N_STATE_ORG;
	pStcCell->LOC_DIFF =  pStcCell->N_STATE_ORG + 1;
	pStcCell->LOC_Vm = 0;
	pStcCell->N_CELL = 0; /* number of cells */
	pStcCell->isAllocated = 0; /* for memory clear */
	pStcCell->iMemFlagCellLoc = 0;
	pStcCell->iMemFlagAlgebraic = 0;
	pStcCell->iMemFlagRate = 0;
	pStcCell->iMemFlagState = 0;
	pStcCell->iMemFlagConst = 0;
	pStcCell->iMemFlagTemp = 0;
	pStcCell->fpInitCellPar = &init_Cell_LuoRudy1CM_device;
	pStcCell->fpInitCellParHost = &init_Cell_LuoRudy1CM_host;
	pStcCell->fpIteration = &iter_Cell_LuoRudy1CM;
	return pStcCell;
}



/* List of functions called for each timestep, currents commented out are only used when modeling pathological conditions */
/********************************************************/

/* Functions that describe the currents begin here */
__host__ __device__ void init_Const_Atomic_LuoRudy1CM (TypeCellML *pConstIn) {
	LuoRudy1CM_Const *pConst;
	pConst = (LuoRudy1CM_Const *)pConstIn;

	pConst->R = 8314.50;      /* Universal Gas Constant (J/kmol*K) */ 
	pConst->frdy = 96485;  /* Faraday's Constant (C/mol) */
	pConst->temp = 310;    /* Temperature (K) */
	pConst->Rtemp = 8314.0 * 310.0;
	pConst->ORtemp = 1.0 / pConst->Rtemp;
	pConst->RtempOfrdy = 8314.0 * 310.0 / 96485.0;
	pConst->frdyORtemp = 96485.0 / (8314.0 * 310.0) ;
	pConst->frdyfrdyORtemp = pConst->frdy * pConst->frdyORtemp;

}


__host__ __device__ void init_State_Atomic_LuoRudy1CM (TypeCellML *pStateIn) {

	LuoRudy1CM_State *pState;

	pState = (LuoRudy1CM_State *)pStateIn;

	pState->Ko = 5.4;
	pState->Nao = 140.0;
	pState->Cao = 1.8;

	pState->CM_xsize = 0.125  ;
	pState->CM_ysize = 0.025 ;
	pState->CM_zsize = 0.025  ;
	pState->CM_C = 0.000125 ; //Capacitance uF
	
	pState->CM_Area = (2*pState->CM_xsize*pState->CM_ysize + 2*pState->CM_ysize*pState->CM_zsize + 2*pState->CM_zsize*pState->CM_xsize)  ; 
	pState->CM_Cm = pState->CM_C/pState->CM_Area ; // 0.09090909
	pState->CM_CmJ = 125.0 ; // pF 

	pState->CM_StimPeriod = 1000.0;
	pState->CM_StimDuration = 1.00;
	pState->CM_StimCurrent = 0.50;
	pState->CM_gNa = 0.16;
	pState->CM_gammaNai = 0.75;
	pState->CM_gammaNao = 0.75;
	pState->CM_gammaKi = 0.75;
	pState->CM_gammaKo = 0.75;
	pState->CM_PCa = 5.4e-6;
	pState->CM_PNa = 6.75e-9;
	pState->CM_PK = 1.93e-9;
	pState->CM_gammaCai = 1.0;
	pState->CM_gammaCao = 0.34;
	pState->CM_Km_Ca = 0.6e-3;
	pState->CM_gK_max = 2.82e-3;
	pState->CM_PR_NaK = 0.01833;
	pState->CM_gK1_max = 7.5e-3;
	pState->CM_gK1_max = 7.5e-3;
	pState->CM_gKp = 1.83e-4;
	pState->CM_K_mpCa = 0.5e-3;
	pState->CM_I_pCa = 1.15e-2;
	pState->CM_gNab = 1.41e-5;
	pState->CM_gCab = 3.016e-5;
	pState->CM_I_NaK = 1.5e-2;
	pState->CM_K_mNai = 10.0;
	pState->CM_K_mKo = 1.5;
	pState->CM_P_ns_Ca = 1.75e-9;
	pState->CM_K_m_ns_Ca = 1.2e-3;
	pState->CM_K_NaCa = 20.0;
	pState->CM_K_mNa = 87.5;
	pState->CM_K_mCa = 1.38;
	pState->CM_K_sat = 0.1;
	pState->CM_eta = 0.35;
	pState->CM_G_relMax = 60.0;
	pState->CM_tau_on = 2.0;
	pState->CM_tau_off = 2.0;
	pState->CM_t_CICR = 0.0;
	pState->CM_tau_tr = 180.0;
	pState->CM_K_mrel = 0.8e-3;
	pState->CM_K_mup = 0.92e-3;
	pState->CM_I_up = 0.005;
	pState->CM_Ca_NSR_max = 15.0;
	pState->CM_DeltaCa_i2 = 0.0;
	pState->CM_DeltaCa_ith = 0.18e-3;
	pState->CM_Am = 200;
	pState->CM_V_myo = 0.68;
	pState->CM_V_JSR = 0.0048;
	pState->CM_V_NSR = 0.0552;
	pState->CM_gK =  pState->CM_gK_max*(pow((pState->Ko/5.40000), 1.0 / 2)) ;
	pState->CM_gK1 =  pState->CM_gK1_max*(pow((pState->Ko/5.40000), 1.0 / 2));
	pState->CM_sigma =  (1.00000/7.00000)*((exp((pState->Nao/67.3000))) - 1.00000);
	pState->CM_G_relPeak = (pState->CM_DeltaCa_i2<pState->CM_DeltaCa_ith ? 0.000000 : pState->CM_G_relMax);
	pState->CM_K_leak = pState->CM_I_up/pState->CM_Ca_NSR_max;
	pState->CM_G_rel =  pState->CM_G_relPeak*((pState->CM_DeltaCa_i2 - pState->CM_DeltaCa_ith)/((pState->CM_K_mrel + pState->CM_DeltaCa_i2) - pState->CM_DeltaCa_ith))*(1.00000 - (exp((- (pState->CM_t_CICR/pState->CM_tau_on)))))*(exp((- (pState->CM_t_CICR/pState->CM_tau_off))));

	pState->CM_IK1_value = 0.5 ;
	pState->CM_ICaL_value = 0.5 ;
	pState->CM_INa_value = 1.00 ;
	pState->CM_IK_value = 1.25 ;
	pState->CM_IKp_value  = 1.00 ;
  
	pState->CM_V = -84.624;
	pState->CM_Nai = 10.0;
	pState->CM_xm = 0.0;
	pState->CM_xh = 1.0;
	pState->CM_xj = 1.0;
	pState->CM_Cai = 0.12e-3;
	pState->CM_Ki = 145.0;
	pState->CM_xd = 0.0;
	pState->CM_xf = 1.0;
	pState->CM_xx = 0.0;
	pState->CM_Ca_JSR = 1.8;
	pState->CM_Ca_NSR = 1.8;
	
	//useless process 
	pState->dCM_V = 0.0;
	pState->dCM_Nai = 0.0;
	pState->dCM_xm = 0.0;
	pState->dCM_xh = 0.0;
	pState->dCM_xj = 0.0;
	pState->dCM_Cai = 0.0;
	pState->dCM_Ki = 0.0;
	pState->dCM_xd = 0.0;
	pState->dCM_xf = 0.0;
	pState->dCM_xx = 0.0;
	pState->dCM_Ca_JSR = 0.0;
	pState->dCM_Ca_NSR = 0.0;

	// stim + diffusion
	pState->CM_I_st = 0.0;
	pState->CM_I_diff = 0.0;
}

__global__ void kernel_init_StateConst_LuoRudy1CM(TypeCellML *pState, TypeCellML *pConst, 
			 		int iSizeState, int iSizeConst, int iTotal) 
{
	int idx;
	TypeCellML *locState;
	TypeCellML *locConst;	
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {	
		locState = pState + idx * iSizeState;
		locConst = pConst + idx * iSizeConst;
		init_State_Atomic_LuoRudy1CM(locState);
		init_Const_Atomic_LuoRudy1CM(locConst);
	}

}



void init_Cell_LuoRudy1CM_host(TypeCellML *pState, TypeCellML *pConst, int iNum) 
{
	int i;
	TypeCellML *pS;
	TypeCellML *pC;
	StcCell *pCell;

	pCell = prototype_Cell_LuoRudy1CM();

	for (i=0;i<iNum;i++) {
		pS = pState + i*pCell->N_STATE;
		pC = pConst + i*pCell->N_CONST;
		init_State_Atomic_LuoRudy1CM(pS);
		init_Const_Atomic_LuoRudy1CM(pC);
	}
	free(pCell);
}


void init_Cell_LuoRudy1CM_device(StcCell *pStcCell)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_init_StateConst_LuoRudy1CM <<< iBlockDim, iCudaBlockSize >>> 
			(pStcCell->d_State, pStcCell->d_Const, 
			 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL);

}




// calculation routines
// individual ion current calculation


__device__ void CM_calculate_temporal_rate(LuoRudy1CM_State * __restrict__ pState, const LuoRudy1CM_Const * __restrict__ pConst, TypeCellML dt)
{	
	//double Istim ;
	TypeCellML FonRT, v, F, RT;	
	FonRT = pConst->frdyORtemp ; // F/(R*T) ;

	F = pConst->frdy;
	RT = pConst->Rtemp;

	v = pState->CM_V;

	// Na current
	if (v <= - 40) {
		pState->CM_alpha_h = 0.135000*(exp(((80.0000+v)/- 6.80000))) ;
		pState->CM_beta_h = 3.56000*(exp(( 0.0790000*v)))+ 310000.*(exp(( 0.350000*v))) ;
	}else {
		pState->CM_alpha_h = 0.00 ;
		pState->CM_beta_h = 1.00000/( 0.130000*(1.00000+(exp(((v+10.6600)/- 11.1000))))) ;
	}
	

	if (v <= - 40) {
		pState->CM_alpha_j = ( ( - 127140.*(exp(( 0.244400*v))) -  3.47400e-05*(exp(( - 0.0439100*v))))*1.00000*(v+37.7800))/(1.00000+(exp(( 0.311000*(v+79.2300)))));
		pState->CM_beta_j = ( 0.121200*(exp(( - 0.0105200*v))))/(1.00000+(exp(( - 0.137800*(v+40.1400)))));
	}else {
		pState->CM_alpha_j = 0.000000;
		pState->CM_beta_j = ( 0.300000*(exp(( - 2.53500e-07*v))))/(1.00000+(exp(( - 0.100000*(v+32.0000))))); 
	}
	
	
	if ((fabs((v+47.1300)))>0.00100000) {
		pState->CM_alpha_m = (0.320000*1.00000*(v + 47.1300))/(1.00000 - (exp(( - 0.100000*(v+47.1300))))) ;
	}else{
		pState->CM_alpha_m = 3.20000 ;
	}
	
	pState->CM_beta_m =  0.0800000*(exp((- v/11.0000)));
	

	// INa current 
	pState->CM_E_Na =  (( RT)/F)*(log((pState->Nao/pState->CM_Nai)));
	pState->CM_i_Na =  pState->CM_gNa*(pow(pState->CM_xm, 3.00000))*pState->CM_xh*pState->CM_xj*(v - pState->CM_E_Na);

	// Time dependent K current X gate 
	pState->CM_alpha_X = ( 7.19000e-005*(v+30.0000))/(1.00000 - (exp(( -0.148000*(v+30.0000)))));
	pState->CM_beta_X = ( 0.000131000*(v+30.0000))/(-1.00000+(exp(( 0.0687000*(v+30.0000)))));
	pState->dCM_xx =  pState->CM_alpha_X*(1.00000 - pState->CM_xx) -  pState->CM_beta_X*pState->CM_xx;
	
	// L-Type Ca channel d gate
	pState->CM_d_infinity = 1.00000/(1.00000+(exp((- ((v+10.0000)/6.24000)))));
	pState->CM_tau_d =  pState->CM_d_infinity*((1.00000 - (exp((- ((v+10.0000)/6.24000)))))/( 0.0350000*(v+10.0000)));
	pState->CM_alpha_d = pState->CM_d_infinity/pState->CM_tau_d;
	pState->CM_beta_d = (1.00000 - pState->CM_d_infinity)/pState->CM_tau_d;
	pState->dCM_xd =  pState->CM_alpha_d*(1.00000 - pState->CM_xd) -  pState->CM_beta_d*pState->CM_xd;
	
	//L-Type Ca channel f gate 
	pState->CM_f_infinity = 1.00000/(1.00000+(exp(((v+35.0600)/8.60000))))+0.600000/(1.00000+(exp(((50.0000 - v)/20.0000))));
	pState->CM_tau_f = 1.00000/( 0.0197000*(exp((-1*(( 0.0337000*(v+10.0000))*( 0.0337000*(v+10.0000))))))+0.0200000);
	pState->CM_alpha_f = pState->CM_f_infinity/pState->CM_tau_f;
	pState->CM_beta_f = (1.00000 - pState->CM_f_infinity)/pState->CM_tau_f;
	pState->dCM_xf =  pState->CM_alpha_f*(1.00000 - pState->CM_xf) -  pState->CM_beta_f*pState->CM_xf;
	
	// Time dependent K current (EK)
	pState->CM_E_K =  (( RT)/F)*(log(((pState->Ko+ pState->CM_PR_NaK*pState->Nao)/(pState->CM_Ki+ pState->CM_PR_NaK*pState->CM_Nai))));
	
	//Time dependent K current Xi gate 
	pState->CM_Xi = 1.00000/(1.00000+(exp(((v - 56.2600)/32.1000))));
	
	// Time dependent K current IK
	pState->CM_i_K =  pState->CM_IK_value*pState->CM_gK*((pState->CM_xx)*(pState->CM_xx))*pState->CM_Xi*(v - pState->CM_E_K);

	// Time independent K current E_K1
	pState->CM_E_K1 =  (( RT)/F)*(log((pState->Ko/pState->CM_Ki)));

	// Time independent K current K1 gate 
	pState->CM_alpha_K1 = 1.02000/(1.00000+(exp(( 0.238500*((v - pState->CM_E_K1) - 59.2150)))));
	pState->CM_beta_K1 = ( 0.491240*(exp(( 0.0803200*((v+5.47600) - pState->CM_E_K1))))+(exp(( 0.0617500*(v - (pState->CM_E_K1+594.310))))))/(1.00000+(exp(( -0.514300*((v - pState->CM_E_K1)+4.75300)))));
	pState->CM_K1_infinity = pState->CM_alpha_K1/(pState->CM_alpha_K1+pState->CM_beta_K1);
	
	//Time independent K1 current (IK1) 
	pState->CM_i_K1 =  pState->CM_IK1_value*pState->CM_gK1*pState->CM_K1_infinity*(v - pState->CM_E_K1);
	
	// Plateau K current 
	pState->CM_E_Kp = pState->CM_E_K1;
	pState->CM_Kp = 1.00000/(1.00000+(exp(((7.48800 - v)/5.98000))));
	
	// Plateau K current IKp
	pState->CM_i_Kp =  pState->CM_IKp_value*pState->CM_gKp*pState->CM_Kp*(v - pState->CM_E_Kp);

	// Na-K pump I_NaK --- check this eq 
	pState->CM_f_NaK = 1.00000/((1.00000+ 0.124500*(exp(( -0.100000*(( v*F)/(RT))))))+ 0.0365000*pState->CM_sigma*(exp((- (( v*F)/( RT))))));
	pState->CM_xi_NaK =  pState->CM_I_NaK*pState->CM_f_NaK*(1.00000/(1.00000+(pow((pState->CM_K_mNai/pState->CM_Nai), 1.50000))))*(pState->Ko/(pState->Ko+pState->CM_K_mKo));

	// L-type Ca current 
	pState->CM_I_CaK =  pState->CM_PK*(pow(1.00000, 2.00000))*(( v*(pow(F, 2.00000)))/( RT))*(( pState->CM_gammaKi*pState->CM_Ki*(exp((( 1.00000*v*F)/( RT)))) -  pState->CM_gammaKo*pState->Ko)/((exp((( 1.00000*v*F)/( RT)))) - 1.00000));
	// L-Type Ca current fCa gate
	pState->CM_f_Ca = 1.00000/(1.00000+(pow((pState->CM_Cai/pState->CM_Km_Ca), 2.00000)));

	// L-type Ca current 
	pState->CM_xi_CaK =  pState->CM_xd*pState->CM_xf*pState->CM_f_Ca*pState->CM_I_CaK;
	
	// Non-specific Ca activated Current
	pState->CM_EnsCa =  (( RT)/F)*(log(((pState->Ko+pState->Nao)/(pState->CM_Ki+pState->CM_Nai))));
	pState->CM_Vns = v - pState->CM_EnsCa;
	pState->CM_I_ns_K =  pState->CM_P_ns_Ca*(pow(1.00000, 2.00000))*(( pState->CM_Vns*(pow(F, 2.00000)))/( RT))*(( pState->CM_gammaKi*pState->CM_Ki*(exp((( 1.00000*pState->CM_Vns*F)/( RT)))) -  pState->CM_gammaKo*pState->Ko)/((exp((( 1.00000*pState->CM_Vns*F)/( RT)))) - 1.00000));
	pState->CM_xi_ns_K =  pState->CM_I_ns_K*(1.00000/(1.00000+(pow((pState->CM_K_m_ns_Ca/pState->CM_Cai), 3.00000))));
	
	// ion concentration (K cyt)
	pState->dCM_Ki =  - (pState->CM_xi_CaK+pState->CM_i_K+pState->CM_i_K1+pState->CM_i_Kp+pState->CM_xi_ns_K - ( pState->CM_xi_NaK*2.00000))*(pState->CM_Am/( pState->CM_V_myo*F));
	
	// I NCX ; Sodium-Ca Exchanger current
	pState->CM_i_NaCa =  pState->CM_K_NaCa*(1.00000/((pow(pState->CM_K_mNa, 3.00000))+(pow(pState->Nao, 3.00000))))*(1.00000/(pState->CM_K_mCa+pState->Cao))*(1.00000/(1.00000+ pState->CM_K_sat*(exp(( (pState->CM_eta - 1.00000)*v*(F/( RT)))))))*( (exp(( pState->CM_eta*v*(F/( RT)))))*(pow(pState->CM_Nai, 3.00000))*pState->Cao -  (exp(( (pState->CM_eta - 1.00000)*v*(F/( RT)))))*(pow(pState->Nao, 3.00000))*pState->CM_Cai);
	
	// Na back ground current
	pState->CM_E_NaN = pState->CM_E_Na;
	pState->CM_i_Na_b =  pState->CM_gNab*(v - pState->CM_E_NaN);

	// L-type Ca current
	pState->CM_I_CaNa =  pState->CM_PNa*(pow(1.00000, 2.00000))*(( v*(pow(F, 2.00000)))/( RT))*(( pState->CM_gammaNai*pState->CM_Nai*(exp((( 1.00000*v*F)/( RT)))) -  pState->CM_gammaNao*pState->Nao)/((exp((( 1.00000*v*F)/( RT)))) - 1.00000));
	pState->CM_xi_CaNa =  pState->CM_xd*pState->CM_xf*pState->CM_f_Ca*pState->CM_I_CaNa;
		
	// non-specitific Current
	pState->CM_I_ns_Na =  pState->CM_P_ns_Ca*(pow(1.00000, 2.00000))*(( pState->CM_Vns*(pow(F, 2.00000)))/( RT))*(( pState->CM_gammaNai*pState->CM_Nai*(exp((( 1.00000*pState->CM_Vns*F)/( RT)))) -  pState->CM_gammaNao*pState->Nao)/((exp((( 1.00000*pState->CM_Vns*F)/( RT)))) - 1.00000));
	pState->CM_xi_ns_Na =  pState->CM_I_ns_Na*(1.00000/(1.00000+(pow((pState->CM_K_m_ns_Ca/pState->CM_Cai), 3.00000))));
	
	// Ion Concentration Na 
	pState->dCM_Nai =  - (pState->CM_i_Na+pState->CM_xi_CaNa+pState->CM_i_Na_b+pState->CM_xi_ns_Na+ pState->CM_i_NaCa*3.00000+ pState->CM_xi_NaK*3.00000)*(pState->CM_Am/( pState->CM_V_myo*F));

	// Stimulation current
	
	//L-Type Ca current component ; check
	pState->CM_I_CaCa =  pState->CM_PCa*(pow(2.00000, 2.00000))*(( v*(pow(F, 2.00000)))/( RT))*(( pState->CM_gammaCai*pState->CM_Cai*(exp((( 2.00000*v*F)/( RT)))) -  pState->CM_gammaCao*pState->Cao)/((exp((( 2.00000*v*F)/( RT)))) - 1.00000));
	pState->CM_xi_CaCa =  pState->CM_xd*pState->CM_xf*pState->CM_f_Ca*pState->CM_I_CaCa;
	pState->CM_i_Ca_L = pState->CM_ICaL_value*pState->CM_xi_CaCa+pState->CM_xi_CaK+pState->CM_xi_CaNa;
	
	// sarcolemmal Ca Pump
	pState->CM_i_p_Ca =  pState->CM_I_pCa*(pState->CM_Cai/(pState->CM_K_mpCa+pState->CM_Cai));
	
	// Ca background current
	pState->CM_E_CaN =  (( RT)/( 2.00000*F))*(log((pState->Cao/pState->CM_Cai)));	
	pState->CM_i_Ca_b =  pState->CM_gCab*(v - pState->CM_E_CaN);

	// non specific Ca current
	pState->CM_i_ns_Ca = pState->CM_xi_ns_Na+pState->CM_xi_ns_K;

	// total current
	
	pState->CM_i_rel =  pState->CM_G_rel*(pState->CM_Ca_JSR - pState->CM_Cai);
	pState->CM_xi_up =  pState->CM_I_up*(pState->CM_Cai/(pState->CM_Cai+pState->CM_K_mup));
	pState->CM_i_leak =  pState->CM_K_leak*pState->CM_Ca_NSR;
	
	pState->dCM_Cai =  - ((pState->CM_xi_CaCa+pState->CM_i_p_Ca+pState->CM_i_Ca_b) - pState->CM_i_NaCa)*(pState->CM_Am/( 2.00000*pState->CM_V_myo*F))+ pState->CM_i_rel*(pState->CM_V_JSR/pState->CM_V_myo)+ (pState->CM_i_leak - pState->CM_xi_up)*(pState->CM_V_NSR/pState->CM_V_myo);
	pState->CM_i_tr = (pState->CM_Ca_NSR - pState->CM_Ca_JSR)/pState->CM_tau_tr;
	pState->dCM_Ca_JSR = - (pState->CM_i_rel -  pState->CM_i_tr*(pState->CM_V_NSR/pState->CM_V_JSR));
	pState->dCM_Ca_NSR = - ((pState->CM_i_leak+pState->CM_i_tr) - pState->CM_xi_up);
}


__device__ void Cells_calculate_temporal_rate(LuoRudy1CM_State * __restrict__ pState, const LuoRudy1CM_Const * __restrict__ pConst, TypeCellML dt){

	TypeCellML v;	
	v = pState->CM_V;

	pState->CM_dV_dt = ( - (pState->CM_i_Na + pState->CM_i_Ca_L + pState->CM_i_K + pState->CM_i_K1 + pState->CM_i_Kp + pState->CM_i_NaCa + pState->CM_i_p_Ca + pState->CM_i_Na_b + pState->CM_i_Ca_b+pState->CM_xi_NaK+pState->CM_i_ns_Ca))/pState->CM_Cm - pState->CM_I_st;

	pState->dCM_V = pState->CM_dV_dt + pState->CM_I_diff/pState->CM_CmJ;
	// add stim & diffusion
}


__device__ void time_evolution(LuoRudy1CM_State * __restrict__ pState, const LuoRudy1CM_Const * __restrict__ pConst, TypeCellML dt) {
	TypeCellML mtau, htau, jtau, mss, hss, jss;
	pState->CM_Nai = pState->CM_Nai + pState->dCM_Nai*dt ;
	// using exact solution for i_Na
	mtau = 1.0/(pState->CM_alpha_m + pState->CM_beta_m);
	htau = 1.0/(pState->CM_alpha_h + pState->CM_beta_h);
	jtau = 1.0/(pState->CM_alpha_j + pState->CM_beta_j);
	mss = pState->CM_alpha_m * mtau;
	hss = pState->CM_alpha_h * htau;
	jss = pState->CM_alpha_j * jtau;
	pState->CM_xm = mss - (mss - pState->CM_xm) 
			*exp(- dt / mtau);
	pState->CM_xh = hss - (hss - pState->CM_xh) 
			* exp(- dt / htau);
	pState->CM_xj = jss - (jss - pState->CM_xj)
			* exp(- dt / jtau);
	// end for i_Na
	pState->CM_Cai = pState->CM_Cai + pState->dCM_Cai*dt;
	pState->CM_Ki = pState->CM_Ki + pState->dCM_Ki*dt ;
	pState->CM_xd = pState->CM_xd + pState->dCM_xd*dt ;
	pState->CM_xf = pState->CM_xf + pState->dCM_xf*dt ;
	pState->CM_xx = pState->CM_xx + pState->dCM_xx*dt ;
	pState->CM_Ca_JSR = pState->CM_Ca_JSR + pState->dCM_Ca_JSR*dt;
	pState->CM_Ca_NSR = pState->CM_Ca_NSR + pState->dCM_Ca_NSR*dt ;

	pState->CM_V = pState->CM_V + pState->dCM_V*dt;
	
}
/************************************************************************************/
// cuda calculation

__global__ void kernel_Compute_LuoRudy1CM(TypeCellML *pStateIn, TypeCellML *pConstIn, 
			 int n_State, int n_Const,  int iTotal, TypeCellML dt)
{
	int idx;

	LuoRudy1CM_State *d_pState;
	LuoRudy1CM_Const *d_pConst;

	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {

	d_pState = (LuoRudy1CM_State *)(pStateIn + idx * n_State);
	d_pConst = (LuoRudy1CM_Const *)(pConstIn + idx * n_Const);

	CM_calculate_temporal_rate (d_pState, d_pConst, dt);
	Cells_calculate_temporal_rate (d_pState, d_pConst, dt);
	time_evolution  (d_pState, d_pConst, dt);
	}
}


void iter_Cell_LuoRudy1CM (StcCell *pStcCell, TypeCellML deltaT)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_Compute_LuoRudy1CM <<< iBlockDim, iCudaBlockSize >>> 
		(pStcCell->d_State, pStcCell->d_Const, 
		 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL,
		 deltaT);

}



