/* MacCannell CF model */
/* Biophysical J 92 2007 4121-4132 */

/*Add Na current in MacCannell CF Model*/

#ifndef Cell_McCannonINaCF_H
#define Cell_McCannonINaCF_H

#include "StcCell.h"

#ifdef __cplusplus

	extern "C" {
#endif

// variables and constants
typedef struct {
	TypeCellML V_f;
	TypeCellML K_o, Na_i;
	TypeCellML E_K, E_Na;
	TypeCellML I_total, I_Kv, I_K1, I_NaK, I_bNa, I_Na, I_tNa; // currents
	TypeCellML r_Kv, s_Kv; // activation inactivation gates for IKv
	TypeCellML a_K1, b_K1; // activation inactivaiton gates for IK1
    TypeCellML xm, xj, xh ; // activation inactivaiton gates for INa
	TypeCellML tm, th, th1, th2, tFNa; // TTX sensitive INa gates
 	TypeCellML I_stim, I_gap; // stimulation and diffusion current
} MacCannellINaCF_State;

typedef struct {
	TypeCellML C_m; // 0
    TypeCellML R, T, F, FonRT;  // 1
	TypeCellML g_Kv; // I_Kv, 5
	TypeCellML g_K1; // I_K1, 6
    TypeCellML g_xNa, g_tNa, Nao; // I_Na, 7
	TypeCellML Ibar_NaK, K_mK, K_mNa, V_rev, B; // INaK, 10
	TypeCellML g_bNa; // backgroun (leak) Na current, 15
} MacCannellINaCF_Const;

typedef union {
	TypeCellML *pDouble;
	MacCannellINaCF_State *pState;
	MacCannellINaCF_Const *pConst;
} MacCannellINaCF_union;

StcCell *prototype_Cell_MacCannellINaCF(void);
int alloc_Cell_MacCannellINaCF_device(StcCell *pStcCell);
void init_Cell_MacCannellINaCF_host(TypeCellML *pState, TypeCellML *pConst, int num);
void init_Cell_MacCannellINaCF_device(StcCell *pStcCell);
void iter_Cell_MacCannellINaCF(StcCell *pStcCell, TypeCellML deltaT);


#ifdef __cplusplus
}
#endif


#endif // define CellML_SachseCF_H
