
#include <stdlib.h>
#include <string.h>
// #include "StcCell.h"
#include "Cell_MacCannellINaCF.h"
#include "math.h"

// #include "StcMultiCell.h"

// parameter initiation


StcCell *prototype_Cell_MacCannellINaCF(void) // StcCell *pStcCell)
{
	StcCell *pStcCell;
	pStcCell = (StcCell *)malloc(sizeof(StcCell));
	pStcCell->N_ALGEBRAIC = 0; // not using
	pStcCell->N_RATE = 0; // not using
	pStcCell->N_STATE_ORG = sizeof(MacCannellINaCF_State)/sizeof(TypeCellML) - 2; 
	pStcCell->N_STATE = pStcCell->N_STATE_ORG + 2; /* additional two for diffusion and stimulation */
	pStcCell->N_CONST = sizeof(MacCannellINaCF_Const)/sizeof(TypeCellML);
	pStcCell->LOC_STIM = pStcCell->N_STATE_ORG;
	pStcCell->LOC_DIFF =  pStcCell->N_STATE_ORG + 1;
	pStcCell->LOC_Vm = 0;
	pStcCell->N_CELL = 0; /* number of cells */
	pStcCell->isAllocated = 0; /* for memory clear */
	pStcCell->iMemFlagCellLoc = 0;
	pStcCell->iMemFlagAlgebraic = 0;
	pStcCell->iMemFlagRate = 0;
	pStcCell->iMemFlagState = 0;
	pStcCell->iMemFlagConst = 0;
	pStcCell->iMemFlagTemp = 0;
	pStcCell->fpInitCellPar = &init_Cell_MacCannellINaCF_device;
	pStcCell->fpInitCellParHost = &init_Cell_MacCannellINaCF_host;
	pStcCell->fpIteration = &iter_Cell_MacCannellINaCF;
	return pStcCell;
}

// number of states and constants
__host__ __device__ void init_Const_Atomic_MacCannellINaCF(TypeCellML *pConstIn) {
	MacCannellINaCF_Const *pConst;
	pConst = (MacCannellINaCF_Const *)pConstIn;
    pConst->R = 8.314472;
    pConst->T = 308;
    pConst->F = 96.4853415;
    pConst->FonRT = pConst->F / (pConst->R * pConst->T);
    pConst->g_Kv = 0.25 ; // nS/pF 
    pConst->g_K1 = 482.2 / 1000.0 ; // pS/pF to nS/pF
    pConst->Ibar_NaK = 2.002 ; 
    pConst->K_mK = 1.0;  // mM/L
    pConst->K_mNa = 11.0; // mM/L
    pConst->V_rev = -150.0 ; // mV
    pConst->B = -200.0; // mV
    pConst->g_bNa = 0.0095 ; // nS/pF
	pConst->g_xNa = 0.60 ; //nS/pF (5% of myocytes)    
	pConst->g_tNa = 0.50 ; // TTX sensitive Na, 2/3 of INa
    pConst->C_m = 6.3; // pF
    pConst->Nao = 140.0; // 140 mM
}


__host__ __device__ void init_State_Atomic_MacCannellINaCF(TypeCellML *pStateIn) {
	   MacCannellINaCF_State *pState;
	   TypeCellML alpha, beta, V_f;
	   pState = (MacCannellINaCF_State *)pStateIn;
	   pState->V_f = -49.6;
	   V_f = pState->V_f;
	   pState->K_o = 5.0 ; // mM
	   pState->Na_i = 10.735 ;
	   pState->E_K = -87.0;
	   pState->E_Na = 8.314472 * 308.0 / 96.4853415 * log(140.0/10.735) ; //40.0; // please update
	   pState->r_Kv = 0.063;
	   pState->s_Kv = 0.978119;
	   pState->a_K1 = 0.0; 
	   pState->b_K1 = 0.0;
	   pState->I_stim = 0.0;
	   pState->I_gap = 0.0;
	   pState->I_Na = 0.0 ;
	   // INa initial
	   alpha = 0.135000*(exp(((80.0000+V_f)/- 6.80000))) ;
	   beta = 3.56000*(exp(( 0.0790000*V_f)))+ 310000.*(exp(( 0.350000*V_f))) ;
	   pState->xh = alpha / (alpha + beta);
	   alpha = ( ( - 127140.*(exp(( 0.244400*V_f))) -  3.47400e-05*(exp(( - 0.0439100*V_f))))*1.00000*(V_f+37.7800))/(1.00000+(exp(( 0.311000*(V_f+79.2300)))));
	   beta = ( 0.121200*(exp(( - 0.0105200*V_f))))/(1.00000+(exp(( - 0.137800*(V_f+40.1400)))));
	   pState->xj = alpha / (alpha + beta);
	   if ((fabs((V_f+47.1300)))>0.00100000) {
		alpha = (0.320000*1.00000*(V_f + 47.1300))/(1.00000 - (exp(( - 0.100000*(V_f + 47.1300))))) ;
	   }else{
		alpha = 3.20000 ;
	   }
	   beta = 0.0800000*(exp((- V_f/11.0000)));
	   pState->xm = alpha / (alpha+beta);

	   // TTX sensitive INa
	   pState->I_tNa = 0.0; 
	   pState->tm = 0.124 ;
	   pState->th1 = 0.595 ;
	   pState->th2 = 5.250e-2 ; 
	   
}


__global__ void kernel_init_StateConst_MacCannellINaCF(TypeCellML *pState, TypeCellML *pConst, 
			 						int iSizeState, int iSizeConst, int iTotal) 
	// , int N_STATE, int N_CONSTANT) 
{
	int idx;
	TypeCellML *locState;
	TypeCellML *locConst;	
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {	
		locState = pState + idx * iSizeState;
		locConst = pConst + idx * iSizeConst;
		init_State_Atomic_MacCannellINaCF(locState);
		init_Const_Atomic_MacCannellINaCF(locConst);
	}

}


void init_Cell_MacCannellINaCF_host(TypeCellML *pState, TypeCellML *pConst, int iNum) 
// , int N_STATE, int N_CONSTANT)
{
	int i;
	TypeCellML *pS;
	TypeCellML *pC;
	StcCell *pCell;

	pCell = prototype_Cell_MacCannellINaCF();

	for (i=0;i<iNum;i++) {
		pS = pState + i*pCell->N_STATE;
		pC = pConst + i*pCell->N_CONST;
		init_State_Atomic_MacCannellINaCF(pS);
		init_Const_Atomic_MacCannellINaCF(pC);
	}
	free(pCell);
}

void init_Cell_MacCannellINaCF_device(StcCell *pStcCell)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_init_StateConst_MacCannellINaCF <<< iBlockDim, iCudaBlockSize >>> 
			(pStcCell->d_State, pStcCell->d_Const, 
			 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL);

}



__device__ void device_ComputeAtomic_MacCannellINaCF(MacCannellINaCF_State * __restrict__ pState, const MacCannellINaCF_Const * __restrict__ pConst, TypeCellML dt) {
	TypeCellML tau_r, tau_s;
	TypeCellML tempVal, tempVal1;
	TypeCellML rbar, sbar;
	TypeCellML V_f, dV_f;
	TypeCellML K_ratio, Na_ratio, Na32, V_ratio;
	TypeCellML ah, bh, am, bm, aj, bj ;
	TypeCellML temptau, tempss;
	
	// I_Kv
	V_f = pState->V_f;
	tempVal = (V_f + 20.0) / 25.9;
	tau_r = 20.3 + 138.0 * exp( -tempVal * tempVal );
	rbar = 1.0 / ( 1.0 + exp( -(V_f + 20.0)/11.0 ) );
	// exact solution instead
	pState->r_Kv = pState->r_Kv - (pState->r_Kv-rbar) * (1.0 - exp(-dt/tau_r)); 

	tempVal = (V_f + 23.0) / 22.7;
	tau_s = 1574.0 + 5268.0 * exp( -tempVal*tempVal );
	sbar = 1.0 / ( 1.0 + exp( (V_f + 23)/7.0 ) );
	pState->s_Kv = pState->s_Kv - (pState->s_Kv-sbar) * (1.0 - exp(-dt/tau_s));
	pState->I_Kv = pConst->g_Kv * pState->r_Kv * pState->s_Kv * (V_f - pState->E_K);

	//I_K1
	tempVal = V_f - pState->E_K;
	pState->a_K1 = 0.1 / ( 1.0 + exp( 0.06 * (tempVal - 200.0) ) );
	pState->b_K1 = (3.0 * exp( 0.0002 * (tempVal+100.0) ) + exp( 0.1 * (tempVal - 10.0) )) 
    					/ (1.0 + exp(-0.5 * tempVal)) ;
	pState->I_K1 = pConst->g_K1 * pState->a_K1 * tempVal / (pState->a_K1 + pState->b_K1) ;

	// I_NaK
	K_ratio = pState->K_o / ( pState->K_o + pConst->K_mK );
	Na32 = pow(pState->Na_i, 1.5);
	Na_ratio = Na32 / ( Na32 + pow(pConst->K_mNa, 1.5) );
	V_ratio = (V_f - pConst->V_rev) / (V_f - pConst->B);	
	pState->I_NaK = pConst->Ibar_NaK * K_ratio * Na_ratio * V_ratio;

	// I_bNa
	pState->I_bNa = pConst->g_bNa * (V_f - pState->E_Na);

	//I_Na,TR
	if (V_f <= - 40.0) {
		ah = 0.135000*(exp(((80.0000+V_f)/- 6.80000))) ;
		bh = 3.56000*(exp(( 0.0790000*V_f)))+ 310000.*(exp(( 0.350000*V_f))) ;
	}else {
		ah = 0 ;
		bh = 1.00000/( 0.130000*(1.00000+(exp(((V_f+10.6600)/- 11.1000))))) ;
	}
	
	temptau = 1.0 / (ah + bh);
    	tempss = ah * temptau;
	pState->xh = tempss - (tempss - pState->xh) * exp( - dt / temptau  );

	if (V_f <= - 40.0) {
		aj = ( ( - 127140.*(exp(( 0.244400*V_f))) -  3.47400e-05*(exp(( - 0.0439100*V_f))))*1.00000*(V_f+37.7800))/(1.00000+(exp(( 0.311000*(V_f+79.2300)))));
		bj = ( 0.121200*(exp(( - 0.0105200*V_f))))/(1.00000+(exp(( - 0.137800*(V_f+40.1400)))));
	}else {
		aj = 0.000000;
		bj = ( 0.300000*(exp(( - 2.53500e-07*V_f))))/(1.00000+(exp(( - 0.100000*(V_f + 32.0000))))); 
	}

	temptau = 1.0 / (aj + bj);
    tempss = aj * temptau;
	pState->xj = tempss - (tempss - pState->xj) * exp( - dt / temptau  );

	if ((fabs((V_f+47.1300)))>0.00100000) {
		am = (0.320000*1.00000*(V_f + 47.1300))/(1.00000 - (exp(( - 0.100000*(V_f + 47.1300))))) ;
	}else{
		am = 3.20000 ;
	}
	bm =  0.0800000*(exp((- V_f/11.0000)));

	temptau = 1.0 / (am + bm);
    tempss = am * temptau;
	pState->xm = tempss - (tempss - pState->xm) * exp( - dt / temptau  );	

	tempVal = pState->xm;
	pState->I_Na = pConst->g_xNa*(pState->xh)*(pState->xj)*tempVal*tempVal*tempVal*(V_f - pState->E_Na);

	// TTX sensitive Na current
	// m
	pState->tFNa = (9.52e-2) * exp((-6.3e-2)*(V_f+34.4)) / (1.0 + 1.65*exp(-2.225*(V_f+63.7))) + 8.69e-2;
	tempss = 1.0/(1.0+exp(-V_f/5.46));
	tempss = pow(tempss, 1.0/3.0);
	temptau = (0.6247e-3) / (0.832*exp(-0.335*(V_f+56.7)) + 0.627*exp(0.082*(V_f+65.01))) + 4.0e-5;
	pState->tm = tempss - (tempss - pState->tm) * exp( - dt / temptau  );
	tempVal = pState->tm;
	// h1
	tempss = 1.0 / (1.0+exp((V_f+66.1)/6.4));
	temptau = (3.717e-6) * exp(-0.2815*(V_f+17.11)) / (1.0 + (3.732e-3)*exp(-0.3426*(V_f+37.76))) + 5.977e-4;
	pState->th1 = tempss - (tempss - pState->th1) * exp( - dt / temptau  );
	// h2
	// use same tempss tempss = 1.0 / (1.0+exp((v_f+66.1)/6.4));
	temptau = (3.186e-8) * exp(-0.6129*(V_f+18.8)) / (1.0 + (7.189e-5)*exp(-0.6683*(V_f+34.07))) + 3.556e-3;
	pState->th2 = tempss - (tempss - pState->th2) * exp( - dt / temptau  );
	// total h
	pState->tFNa = (9.52e-2)*exp((-6.3e-2)*(V_f+34.4)) / (1. + 1.66*exp(-0.225*(V_f+63.7))) + 8.69e-2;	
	pState->th = (1.0 - pState->tFNa) * pState->th1 + pState->tFNa * pState->th2;
	// right side
	// if ((V_f > 0.001) || (V_f < -0.001)) {
	if (fabs(V_f) > 0.001) {
		tempVal1 =  (exp((V_f-pState->E_Na)*pConst->FonRT) - 1.0) / (exp(V_f*pConst->FonRT)-1.0);
	} else {
	       	tempVal1 = exp((-pState->E_Na)*pConst->FonRT); // 1.0 ;
	}	
	pState->I_tNa = pConst->g_tNa * tempVal*tempVal*tempVal * pState->th * pConst->Nao * pConst->FonRT * pConst->F * tempVal1 * V_f;


	// I_total
	pState->I_total = pState->I_Kv + pState->I_K1 + pState->I_NaK + pState->I_bNa + pState->I_Na + pState->I_tNa;

	// calculate V;
	dV_f = (- pState->I_total + pState->I_stim + (pState->I_gap / pConst->C_m) ) * dt; 
	pState->V_f = V_f + dV_f;
}


__global__ void kernel_Compute_MacCannellINaCF(TypeCellML *pStateIn, TypeCellML *pConstIn, 
			 int n_State, int n_Const,  int iTotal, TypeCellML dt)
{
	int idx;	
	TypeCellML *pLocState;
	TypeCellML *pLocConst;
	MacCannellINaCF_State *pStcState;
	MacCannellINaCF_Const *pStcConst;
	// MacCannellINaCF_union uStr;
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		pLocConst = pConstIn + idx * n_Const;
		pLocState = pStateIn + idx * n_State;
		pStcConst = (MacCannellINaCF_Const *) pLocConst;
		pStcState = (MacCannellINaCF_State *) pLocState;
		device_ComputeAtomic_MacCannellINaCF(pStcState, pStcConst, dt);
	}
}



void iter_Cell_MacCannellINaCF(StcCell *pStcCell, TypeCellML deltaT)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_Compute_MacCannellINaCF <<< iBlockDim, iCudaBlockSize >>> 
		(pStcCell->d_State, pStcCell->d_Const, 
		 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL,
		 deltaT);

}



