
#include <stdlib.h>
#include <string.h>
#include "Cell_PassiveCF.h"
#include "math.h"

// parameter initiation

StcCell *prototype_Cell_PassiveCF(void) // StcCell *pStcCell)
{
	StcCell *pStcCell;
	pStcCell = (StcCell *)malloc(sizeof(StcCell));
	pStcCell->N_ALGEBRAIC = 0; // not using
	pStcCell->N_RATE = 0; // not using
	pStcCell->N_STATE_ORG = sizeof(PassiveCF_State)/sizeof(TypeCellML) - 2; 
	pStcCell->N_STATE = pStcCell->N_STATE_ORG + 2; /* additional two for diffusion and stimulation */
	pStcCell->N_CONST = sizeof(PassiveCF_Const)/sizeof(TypeCellML);
	pStcCell->LOC_STIM = pStcCell->N_STATE_ORG;
	pStcCell->LOC_DIFF =  pStcCell->N_STATE_ORG + 1;
	pStcCell->LOC_Vm = 0;
	pStcCell->N_CELL = 0; /* number of cells */
	pStcCell->isAllocated = 0; /* for memory clear */
	pStcCell->iMemFlagCellLoc = 0;
	pStcCell->iMemFlagAlgebraic = 0;
	pStcCell->iMemFlagRate = 0;
	pStcCell->iMemFlagState = 0;
	pStcCell->iMemFlagConst = 0;
	pStcCell->iMemFlagTemp = 0;
	pStcCell->fpInitCellPar = &init_Cell_PassiveCF_device;
	pStcCell->fpInitCellParHost = &init_Cell_PassiveCF_host;
	pStcCell->fpIteration = &iter_Cell_PassiveCF;
	return pStcCell;
}


	


// number of states and constants

__host__ __device__ void init_Const_Atomic_PassiveCF(TypeCellML *pConstIn) {
	PassiveCF_Const *pConst;
	pConst = (PassiveCF_Const *)pConstIn;
    	pConst->C_m = 6.3; // pF
	pConst->g_Pass = 0.025;
}


__host__ __device__ void init_State_Atomic_PassiveCF(TypeCellML *pStateIn) {
	   PassiveCF_State *pState;
	   pState = (PassiveCF_State *)pStateIn;
	   pState->V_f = -49.0;
	   pState->E_rev = -49.0;
	   pState->I_stim = 0.0;
	   pState->I_gap = 0.0;
}


__global__ void kernel_init_StateConst_PassiveCF(TypeCellML *pState, TypeCellML *pConst, int iSizeState, int iSizeConst, int iTotal) 
{
	int idx;
	TypeCellML *locState;
	TypeCellML *locConst;	
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {	
		locState = pState + idx * iSizeState;
		locConst = pConst + idx * iSizeConst;
		init_State_Atomic_PassiveCF(locState);
		init_Const_Atomic_PassiveCF(locConst);
	}

}


void init_Cell_PassiveCF_host(TypeCellML *pState, TypeCellML *pConst, int iNum) 
{
	int i;
	TypeCellML *pS;
	TypeCellML *pC;
	StcCell *pCell;

	pCell = prototype_Cell_PassiveCF();

	for (i=0;i<iNum;i++) {
		pS = pState + i*pCell->N_STATE;
		pC = pConst + i*pCell->N_CONST;
		init_State_Atomic_PassiveCF(pS);
		init_Const_Atomic_PassiveCF(pC);
	}
	free(pCell);
}

void init_Cell_PassiveCF_device(StcCell *pStcCell)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_init_StateConst_PassiveCF <<< iBlockDim, iCudaBlockSize >>> 
			(pStcCell->d_State, pStcCell->d_Const, 
			 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL);

}



__device__ void device_ComputeAtomic_PassiveCF(PassiveCF_State *pState, PassiveCF_Const *pConst, TypeCellML dt) {

	TypeCellML V_f, dV_f;

	V_f = pState->V_f;

	pState->I_pass = pConst->g_Pass * (V_f - pState->E_rev);

	// I_total
	pState->I_total = pState->I_pass; 

	// calculate V;
	dV_f = (((- pState->I_total + pState->I_gap) / pConst->C_m) + pState->I_stim) * dt; 
	pState->V_f = V_f + dV_f;
}


__global__ void kernel_Compute_PassiveCF(TypeCellML *pStateIn, TypeCellML *pConstIn, 
			 int n_State, int n_Const,  int iTotal, TypeCellML dt)
{
	int idx;	
	TypeCellML *pLocState;
	TypeCellML *pLocConst;
	PassiveCF_State *pStcState;
	PassiveCF_Const *pStcConst;

	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		pLocConst = pConstIn + idx * n_Const;
		pLocState = pStateIn + idx * n_State;
		pStcConst = (PassiveCF_Const *) pLocConst;
		pStcState = (PassiveCF_State *) pLocState;
		device_ComputeAtomic_PassiveCF(pStcState, pStcConst, dt);
	}
}



void iter_Cell_PassiveCF(StcCell *pStcCell, TypeCellML deltaT)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_Compute_PassiveCF <<< iBlockDim, iCudaBlockSize >>> 
		(pStcCell->d_State, pStcCell->d_Const, 
		 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL,
		 deltaT);

}



