
#include <stdlib.h>
#include <string.h>
#include "Cell_PassiveIVCF.h"
#include "math.h"



// parameter initiation


StcCell *prototype_Cell_PassiveIVCF(void) // StcCell *pStcCell)
{
	StcCell *pStcCell;
	pStcCell = (StcCell *)malloc(sizeof(StcCell));
	pStcCell->N_ALGEBRAIC = 0; // not using
	pStcCell->N_RATE = 0; // not using
	pStcCell->N_STATE_ORG = sizeof(PassiveIVCF_State)/sizeof(TypeCellML) - 2; 
	pStcCell->N_STATE = pStcCell->N_STATE_ORG + 2; /* additional two for diffusion and stimulation */
	pStcCell->N_CONST = sizeof(PassiveIVCF_Const)/sizeof(TypeCellML);
	pStcCell->LOC_STIM = pStcCell->N_STATE_ORG;
	pStcCell->LOC_DIFF =  pStcCell->N_STATE_ORG + 1;
	pStcCell->LOC_Vm = 0;
	pStcCell->N_CELL = 0; /* number of cells */
	pStcCell->isAllocated = 0; /* for memory clear */
	pStcCell->iMemFlagCellLoc = 0;
	pStcCell->iMemFlagAlgebraic = 0;
	pStcCell->iMemFlagRate = 0;
	pStcCell->iMemFlagState = 0;
	pStcCell->iMemFlagConst = 0;
	pStcCell->iMemFlagTemp = 0;
	pStcCell->fpInitCellPar = &init_Cell_PassiveIVCF_device;
	pStcCell->fpInitCellParHost = &init_Cell_PassiveIVCF_host;
	pStcCell->fpIteration = &iter_Cell_PassiveIVCF;
	return pStcCell;
}


	


// number of states and constants


__host__ __device__ void init_Const_Atomic_PassiveIVCF(TypeCellML *pConstIn) {
	PassiveIVCF_Const *pConst;
	pConst = (PassiveIVCF_Const *)pConstIn;
    	pConst->C_m = 6.3; // pF
	pConst->g_Pass = 0.025;
}


__host__ __device__ void init_State_Atomic_PassiveIVCF(TypeCellML *pStateIn) {
	   PassiveIVCF_State *pState;
	   pState = (PassiveIVCF_State *)pStateIn;
	   pState->V_f = -49.0;
	   pState->E_rev = -49.0;
	   pState->I_stim = 0.0;
	   pState->I_gap = 0.0;
}


__global__ void kernel_init_StateConst_PassiveIVCF(TypeCellML *pState, TypeCellML *pConst, int iSizeState, int iSizeConst, int iTotal) 
{
	int idx;
	TypeCellML *locState;
	TypeCellML *locConst;	
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {	
		locState = pState + idx * iSizeState;
		locConst = pConst + idx * iSizeConst;
		init_State_Atomic_PassiveIVCF(locState);
		init_Const_Atomic_PassiveIVCF(locConst);
	}

}


void init_Cell_PassiveIVCF_host(TypeCellML *pState, TypeCellML *pConst, int iNum) 
{
	int i;
	TypeCellML *pS;
	TypeCellML *pC;
	StcCell *pCell;

	pCell = prototype_Cell_PassiveIVCF();

	for (i=0;i<iNum;i++) {
		pS = pState + i*pCell->N_STATE;
		pC = pConst + i*pCell->N_CONST;
		init_State_Atomic_PassiveIVCF(pS);
		init_Const_Atomic_PassiveIVCF(pC);
	}
	free(pCell);
}

void init_Cell_PassiveIVCF_device(StcCell *pStcCell)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_init_StateConst_PassiveIVCF <<< iBlockDim, iCudaBlockSize >>> 
			(pStcCell->d_State, pStcCell->d_Const, 
			 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL);

}



__device__ void device_ComputeAtomic_PassiveIVCF(PassiveIVCF_State *pState, PassiveIVCF_Const *pConst, TypeCellML dt) {
	TypeCellML V_f, dV_f;

	V_f = pState->V_f;

	// grand et al
	pState->I_pass = -12.5 - 0.03 * V_f + exp( (V_f+642.5)/250.0 );

	// I_total
	pState->I_total = pState->I_pass; 
	
	// calculate V;
	dV_f =( (- pState->I_total + pState->I_gap / pConst->C_m) + pState->I_stim) * dt; 
	pState->V_f = V_f + dV_f;
}


__global__ void kernel_Compute_PassiveIVCF(TypeCellML *pStateIn, TypeCellML *pConstIn, 
			 int n_State, int n_Const,  int iTotal, TypeCellML dt)
{
	int idx;	
	TypeCellML *pLocState;
	TypeCellML *pLocConst;
	PassiveIVCF_State *pStcState;
	PassiveIVCF_Const *pStcConst;
	
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		pLocConst = pConstIn + idx * n_Const;
		pLocState = pStateIn + idx * n_State;
		pStcConst = (PassiveIVCF_Const *) pLocConst;
		pStcState = (PassiveIVCF_State *) pLocState;
		device_ComputeAtomic_PassiveIVCF(pStcState, pStcConst, dt);
	}
}



void iter_Cell_PassiveIVCF(StcCell *pStcCell, TypeCellML deltaT)
{
	int iBlockDim;
	iBlockDim = (pStcCell->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStcCell->N_CELL) iBlockDim ++;
	kernel_Compute_PassiveIVCF <<< iBlockDim, iCudaBlockSize >>> 
		(pStcCell->d_State, pStcCell->d_Const, 
		 pStcCell->N_STATE, pStcCell->N_CONST, pStcCell->N_CELL,
		 deltaT);

}



