; sequence of simulation
; need to compile StimLogicFunctions.pro first
; compile ClsStcCell.pro as well
; o = OBJ_NEW('ClsStcCell')
; o->initObj
; o->allocCellsOnMemory
; o->allocCellsOnCuda
; makeConstants
; o->copyConstToDevice
; o->runStimRecord
; o->writeNetCDF, 'test.nc'
; o->setValue, 'StimLogic', 'S20.350.2'
; o->runStimRecord
; .... etc

PRO clearDevice
  ;; set for dll file location.
  dllName = 'C:\MultiCellCuda\cudaIDLCellMLcomm-x64.dll'

  res = CALL_EXTERNAL(dllName, 'IDL_free_diffusion')
  res = CALL_EXTERNAL(dllName, 'IDL_free_ctn', /UNLOAD)
  Print, 'device is cleared'
END



PRO ClsStcCell__DEFINE
  str = {ClsStcCell, $
    iCellType:0L, $
    iCellId:0L, $
    N_STATE:0L, $
    N_CONST:0L, $
    N_CELL:0L, $
    ptrState:PTR_NEW(), $
    ptrConst:PTR_NEW(), $
    ptrCellLoc:PTR_NEW() $
  }
END

PRO ClsStcCell::freePtr
  PTR_FREE, self.ptrState
  PTR_FREE, self.ptrConst
  PTR_FREE, self.ptrCellLoc
END

PRO ClsStcCell::CLEANUP
  self->freePtr
END

PRO ClsStcCell::allocMemoryHost
  self->freePtr
  self.ptrState = PTR_NEW(DBLARR(self.N_STATE, self.N_CELL))
  self.ptrConst = PTR_NEW(DBLARR(self.N_CONST, self.N_CELL))
  self.ptrCellLoc = PTR_NEW(LONARR(self.N_CELL))
END

PRO ClsStcCell::getValue, strOpt, val
  CASE strOpt OF
    'STATE': val = *(self.ptrState)
    'CONST': val = *(self.ptrConst)
    'LOC': val = *(self.ptrCellLoc)
    'INFO': val = [self.N_CELL, self.N_STATE, self.N_CONST]
  ENDCASE
END

PRO ClsDiff__DEFINE
  str = { ClsDiff, $
    iFlagDevice:0, $
    N_CELL:0L, $
    N_NEIGHBOR:0L, $
    pNeighborInd:PTR_NEW(), $
    pDiffCoeff:PTR_NEW()$
  }
END

PRO ClsDiff::CLEANUP
  self->freePtr
END

PRO ClsDiff::getValue, opt, val
  CASE opt OF
    'N_CELL': val = N_CELL
    'N_NEIGHBOR': val = N_NEIGHBOR
    'NeighborInd': val = *(self.pNeighborInd)
    'DiffCoeff': val = *(self.pDiffCoeff)
  ENDCASE
END


PRO ClsDiff::freePtr
  PTR_FREE, self.pNeighborInd
  PTR_FREE, self.pDiffCoeff
END

PRO ClsStcCellCltn__DEFINE
  str = {ClsStcCellCltn, $
    iFlagCltnDevice:0L, $
    iFlagDiffDevice:0L, $
    strDirPath:'', $
    oCltn:OBJ_NEW(), $
    oDiff:OBJ_NEW(), $
    iDim:0L, $
    iDimXY:[0L, 0L], $
    iDimXYZ:[0L, 0L, 0L], $
    N_STCCELL:0L, $
    N_CELL:0L, $
    dt:0.0d, $
    ptrTempData:PTR_NEW(), $
    dllName:'' $
  }
END


PRO ClsStcCellCltn::initObj
  self.N_STCCELL = 0L
  self.N_Cell = 0L
  self.iFlagCltnDevice = 0L
  self.dt = 0.01d
  self.sDiff = 1L
  self.oCltn = OBJ_NEW('IDL_Container')
  self.oDiff = OBJ_NEW('ClsDiff')
  self.dllName = 'C:\MultiCellCuda\cudaIDLCellMLcomm-x64.dll'
  self->allocCltnDevice
END

PRO ClsStcCellCltn::freeDevice
  IF self.iFlagDiffDevice THEN BEGIN
    res = CALL_EXTERNAL(self.dllName, 'IDL_free_diffusion')
    self.iFlagDiffDevice = 0
  ENDIF
  IF (self.iFlagCltnDevice) THEN BEGIN
    res = CALL_EXTERNAL(self.dllName, 'IDL_free_ctn', /UNLOAD)
    self.iFlagCltnDevice = 0
  ENDIF
END

PRO ClsStcCellCltn::allocCltnDevice
  IF (self.iFlagCltnDevice EQ 0) THEN BEGIN
    res = CALL_EXTERNAL(self.dllName, 'IDL_alloc_ctn')
    self.iFlagCltnDevice = 1
  ENDIF
END


PRO ClsStcCellCltn::setDiffParm, iNeighbor, dDiffCoeff
  iSize = SIZE(iNeighbor)
  self.oDiff.N_NEIGHBOR = iSize[1]
  self.oDiff.N_CELL = iSize[2]
  PTR_FREE, self.oDiff.pNeighborInd
  self.oDiff.pNeighborInd = PTR_NEW(iNeighbor)
  PTR_FREE, self.oDiff.pDiffCoeff
  self.oDiff.pDiffCoeff = PTR_NEW(dDiffCoeff)
END

PRO ClsStcCellCltn::multiplyDiffParm, r
  bm = MEAN(*(self.oDiff.pDiffCoeff))
  *(self.oDiff.pDiffCoeff) *= r
  am = MEAN(*(self.oDiff.pDiffCoeff))
  PRINT, bm, am
END

PRO ClsStcCellCltn::updateDiffParmDevice
  iNeighborInd = *(self.oDiff.pNeighborInd)
  dDiffCoeff = *(self.oDiff.pDiffCoeff)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_DiffParm', iNeighborInd, dDiffCoeff)
End


PRO ClsStcCellCltn::allocDiffDevice
  res = CALL_EXTERNAL(self.dllName, 'IDL_alloc_diffusion', self.oDiff.N_NEIGHBOR)
  iNeighborInd = *(self.oDiff.pNeighborInd)
  dDiffCoeff = *(self.oDiff.pDiffCoeff)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_DiffParm', iNeighborInd, dDiffCoeff)
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(self.dllName, 'IDL_get_info_diffusion', ctninfo)
End

; -1 : if cltn has already same object
; -2 : if obj is different type
; 1  : OK

PRO ClsStcCellCltn::addCell, oCell
  self.oCltn->add, oCell
  self.N_STCCELL = self.oCltn->count()
  self.N_CELL += oCell.N_CELL
  PTR_FREE, self.ptrTempData
  self.ptrTempData = PTR_NEW(DBLARR(self.N_CELL))
END

PRO ClsStcCellCltn::allocCellDevice
  FOR i=0, self.N_STCCELL-1 DO BEGIN
    o = self.oCltn->get(POSITION=I)
    iCellType = o.iCellType
    n_cell = o.n_cell
    res = CALL_EXTERNAL(self.dllname, 'IDL_add_StcCell', LONG(iCellType), LONG(n_cell))
    Print, 'celltype, nc_cell', iCellType, n_cell
    iStrCell = LONG(I+1) ; adjust +1
    res = CALL_EXTERNAL(self.dllName, 'IDL_copy_state_ToDevice', iStrCell, *(o.ptrState))
    res = CALL_EXTERNAL(self.dllName, 'IDL_copy_const_ToDevice', iStrCell, *(o.ptrConst))
    res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellLoc', iStrCell, *(o.ptrCellLoc))
    Print, 'Vm', MEAN( (*(o.ptrState))[0,*])
  ENDFOR

END




PRO ClsStcCellCltn::emptyCltn
  OBJ_DESTROY, self.oCltn
  OBJ_DESTROY, self.oDiff
  self->initObj
END

PRO ClsStcCellCltn::CLEANUP
  self->freeDevice
  Print, 'Device mem freed'
  OBJ_DESTROY, self.oCltn
  OBJ_DESTROY, self.oDiff
END


PRO ClsStcCellCltn::readCells, filename
  IF (fileName EQ '') THEN RETURN
  IF (~FILE_TEST(fileName, /REGULAR)) THEN RETURN
  finfo = FILE_INFO(fileName)
  self->freeDevice
  self->emptyCltn
  id = NCDF_OPEN(fileName, /NOWRITE)

  NCDF_ATTGET, id, /GLOBAL, 'N_STCCELL', n_stccell
  NCDF_ATTGET, id, /GLOBAL, 'N_CELL', n_cell
  NCDF_ATTGET, id, /GLOBAL, 'DIM', iDim
  self.iDim = iDim
  NCDF_ATTGET, id, /GLOBAL, 'DIMXY', iDimXY
  self.iDimXY = iDimXY
  NCDF_ATTGET, id, /GLOBAL, 'DIMXYZ', iDimXYZ
  self.iDimXYZ = iDimXYZ
  Print, self.n_stccell, self.n_cell, self.iDim, self.iDimXY, self.iDimXYZ
  cellTypeArr = LONARR(n_cell)
  FOR I=0, n_stccell-1 DO BEGIN
    oObj = OBJ_NEW('ClsStcCell')
    strI = STRTRIM(string(I),1)
    cellvars = 'CellVar'+ strI
    NCDF_ATTGET, id, /GLOBAL,cellvars, var
    oObj.iCellType = var[0]
    oObj.N_STATE = var[1]
    oObj.N_CONST = var[2]
    oObj.N_CELL = var[3]
    statename = 'STATE'+strI
    NCDF_VARGET, id, statename, MyoState ; states
    oObj.ptrState = PTR_NEW(MyoState)
    constname = 'CONST'+ strI
    NCDF_VARGET, id, constname, MyoConst ; const
    Print, CellVars
    Print, MyoState[*,0]
    Print, MyoConst[*,0]
    oObj.ptrConst = PTR_NEW(MyoConst)
    locname = 'LOC'+strI
    NCDF_VARGET, id, locname, MyoLoc
    ;Help, MyoLoc
    oObj.ptrCellLoc = PTR_NEW(MyoLoc)
    self->addCell, oObj
    cellTypeArr[*(oObj.ptrCellLoc)] = I+1
  ENDFOR
  ; MyoConst = const
  NCDF_VARGET, id, 'NEIGHBOR', neighbor
  n = SIZE(neighbor)
  self.oDiff.N_CELL = n[2]
  self.oDiff.N_NEIGHBOR = n[1]
  self.oDiff.pNeighborInd = PTR_NEW(neighbor)
  NCDF_VARGET, id, 'DIFFCOEFF', coeff
  self.oDiff.pDiffCoeff = PTR_NEW(coeff)
  NCDF_CLOSE, id
 
  iNeighborInd = *(self.oDiff.pNeighborInd)
  couplingArr = *(self.oDiff.pDiffCoeff)
  self->allocCellDevice
  self->allocDiffDevice
 
   ; set time
  Print, 'integration time step', self.dt
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_dt', self.dt)
END


PRO ClsStcCellCltn::writeCells, filename ;, SAVEPATHINCLUDED=savepathincluded
  ; write a file
  ;    IF (KEYWORD_SET(savepathincluded)) THEN fsave = filename $
  ;	    ELSE fsave = self.ssavepath + filename
  fsave = filename
  IF (fileName EQ '') THEN RETURN
  ;self->copyCellsToHost
  id = NCDF_CREATE(fsave, /CLOBBER)
  ; Fill the file with default values:
  NCDF_CONTROL, id, /FILL
  NCDF_ATTPUT, id, /GLOBAL, 'N_STCCELL', self.N_STCCELL
  NCDF_ATTPUT, id, /GLOBAL, 'N_CELL', self.N_CELL
  NCDF_ATTPUT, id, /GLOBAL, 'DIM',  self.iDim
  NCDF_ATTPUT, id, /GLOBAL, 'DIMXY', self.iDimXY
  NCDF_ATTPUT, id, /GLOBAL, 'DIMXYZ', self.iDimXYZ
  
  
  stateid = INTARR(self.N_STCCELL)
  constid = INTARR(self.N_STCCELL)
  celllocid = INTARR(self.N_STCCELL)
  FOR I=0, self.N_STCCELL-1 DO BEGIN
    oObj = self.oCltn->get(POSITION=I)
    strI = STRTRIM(string(I),1)
    varname = 'CellVar' + strI
    arr = [oObj.iCellType, oObj.N_STATE, oObj.N_CONST, oObj.N_CELL]
    NCDF_ATTPUT, id, /GLOBAL, varname, arr
    y = 'y'+ STRTRIM(STRING(I),1)
    ncellid = NCDF_DIMDEF(id, Y, oObj.N_CELL)
    xstate = 'xstate' + strI
    statesizeid = NCDF_DIMDEF(id, xstate, oObj.N_STATE)
    xconst = 'xconst' + strI
    constsizeid = NCDF_DIMDEF(id, xconst, oObj.N_CONST)
    ; Define variables:
    statename = 'STATE' + strI
    constname = 'CONST' + strI
    celllocname = 'LOC' + strI
    stateid[I] = NCDF_VARDEF(id, statename, [statesizeid, ncellid], /DOUBLE)
    constid[I] = NCDF_VARDEF(id, constname, [constsizeid, ncellid], /DOUBLE)
    celllocid[I] = NCDF_VARDEF(id, celllocname, [ncellid], /LONG)
  ENDFOR
  ; diffusion parm
  neighborsizeid = NCDF_DIMDEF(id, 'N_NEIGHBOR', self.oDiff.N_NEIGHBOR)
  ncelltotalid = NCDF_DIMDEF(id, 'N_CELL_TOTAL', self.N_CELL)
  neighborid = NCDF_VARDEF(id, 'NEIGHBOR', [neighborsizeid, ncelltotalid], /LONG)
  diffcoeffid = NCDF_VARDEF(id, 'DIFFCOEFF', [neighborsizeid, ncelltotalid], /DOUBLE)
  NCDF_CONTROL, id, /ENDEF
  ; put data:
  FOR I=0, self.N_STCCELL-1 DO BEGIN
    self->copyCellFromDevice, I
    oObj = self.oCltn->get(POSITION=I)
    oObj->getValue, 'STATE', val
    Print, 'Vm', MEAN( val[0,*])
    NCDF_VARPUT, id, stateid[I], val ; REFORM(MyoState, self.iStates, self.iNCells)
    oObj->getValue, 'CONST', val
    NCDF_VARPUT, id, constid[I], val
    oObj->getValue, 'LOC', val
    NCDF_VARPUT, id, celllocid[I], val
   
   

  ENDFOR
  NCDF_VARPUT, id, neighborid, REFORM(*(self.oDiff.pNeighborInd), self.oDiff.N_NEIGHBOR, self.N_CELL)
  NCDF_VARPUT, id, diffcoeffid, REFORM(*(self.oDiff.pDiffCoeff), self.oDiff.N_NEIGHBOR, self.N_CELL)
  NCDF_CLOSE, id ; Close the NetCDF file
END


PRO ClsStcCellCltn::addCellTypeOnDevice, strCellType, N_CELL
  self->allocCltnDevice
  o = OBJ_NEW('ClsStcCell')
  o.N_CELL = N_CELL
  CASE strCellType OF
    ;		'MahajanCM': iCellType = 0L
    ;		'PanditCM': iCellType = 1L
    ;		'MacCannellCF': icellType = 2L
    'MacCannellINaCF': iCellType = 3L
    'PassiveCF': iCellType = 4L
    ;		'LuoRudyCM': iCellType = 5L
    'LuoRudy1CM': iCellType = 6L
    'PassiveIVCF': iCellType = 7L
  ENDCASE
  o.iCellType = iCellType
  res = CALL_EXTERNAL(self.dllname, 'IDL_add_StcCell', iCellType, LONG(n_cell))
  strinfo = LONARR(5)
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_ctn', strinfo)
  iStrCell = strinfo[1]
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_cell', iStrCell, strinfo)
  o.N_CELL = strinfo[0]
  o.N_STATE = strinfo[1]
  o.N_CONST = strinfo[2]
  o->allocMemoryHost
  self->addCell, o
  res = CALL_EXTERNAL(self.dllName, 'IDL_copy_state_ToHost', iStrCell, *(o.ptrState))
  res = CALL_EXTERNAL(self.dllName, 'IDL_copy_const_ToHost', iStrCell, *(o.ptrConst))
END

PRO ClsStcCellCltn::setCellLoc, cellId, CellLoc
  oObj = self.oCltn->get(POSITION=cellId-1)
  PTR_FREE, oObj.ptrCellLoc
  oObj.ptrCellLoc = PTR_NEW(CellLoc)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellLoc', LONG(CellId), LONG(CellLoc))
  Help, CellId, CellLoc
END

PRO ClsStcCellCltn::copyCellFromDevice, iInd
  strinfo = LONARR(5)
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_ctn', strinfo)
  adjInd = LONG(iInd + 1)
  IF (adjInd GT strinfo[1]) THEN BEGIN
    PRINT, 'No cell found'
    RETURN
  ENDIF
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_cell', adjInd, strinfo)
  o = self.oCltn->get(POSITION=iInd)
  IF o.N_CELL NE strInfo[0] THEN BEGIN
    PRINT, 'N_CELL mismatch'
    RETURN
  ENDIF
  IF o.N_STATE NE strInfo[1] THEN BEGIN
    PRINT, 'N_STATE mismatch'
    RETURN
  ENDIF
  IF (o.N_CONST NE strInfo[2]) THEN BEGIN
    PRINT, 'N_CONST mismatch'
    RETURN
  ENDIF
  iStrCell = adjInd
  res = CALL_EXTERNAL(self.dllName, 'IDL_copy_state_ToHost', iStrCell, *(o.ptrState))
  res = CALL_EXTERNAL(self.dllName, 'IDL_copy_const_ToHost', iStrCell, *(o.ptrConst))
  res = CALL_EXTERNAL(self.dllName, 'IDL_get_CellLoc', iStrCell, *(o.ptrCellLoc))
END

PRO ClsStcCellCltn::copyCellToDevice, iInd
  strinfo = LONARR(5)
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_ctn', strinfo)
  adjInd = LONG(iInd + 1)
  IF (adjInd GT strinfo[1]) THEN BEGIN
    PRINT, 'No cell found'
    RETURN
  ENDIF
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_cell', adjInd, strinfo)
  o = self.oCltn->get(POSITION=adjInd-1)
  IF o.N_CELL NE strInfo[0] THEN BEGIN
    PRINT, 'N_CELL mismatch'
    RETURN
  ENDIF
  IF o.N_STATE NE strInfo[1] THEN BEGIN
    PRINT, 'N_STATE mismatch'
    RETURN
  ENDIF
  IF (o.N_CONST NE strInfo[2]) THEN BEGIN
    PRINT, 'N_CONST mismatch'
    RETURN
  ENDIF
  iStrCell = adjInd
  res = CALL_EXTERNAL(self.dllName, 'IDL_copy_state_ToDevice', iStrCell, *(o.ptrState))
  ;Print, res
  res = CALL_EXTERNAL(self.dllName, 'IDL_copy_const_ToDevice', iStrCell, *(o.ptrConst))
  ;Print, res
END

PRO ClsStcCellCltn::copyAllStateConstToDevice
  FOR i=0, self.N_STCCELL-1 DO BEGIN
    self->copyCellToDevice, I
  ENDFOR
END

PRO ClsStcCellCltn::getVmFromDevice
  res = CALL_EXTERNAL(self.dllName, 'IDL_getVm', *(self.ptrTempData))
END

PRO ClsStcCellCltn::getValuesFromDevice, indArr, varType
  res = CALL_EXTERNAL(self.dllName, 'IDL_getValueSorted', *(self.ptrTempData),indArr, varType)
END

; =========

PRO ClsMultiCellCuda__DEFINE
  str = {ClsMultiCellCuda, $
    t: 0.0d, $ ; simulation time
    dScanInt:0.0d, $ ; sampling time
    iIter: 0L, $; iteration number
    iTotalIter: 0L, $
    strStimLogic: '', $ ; stimulation logic
    pStimArr: PTR_NEW(), $
    iStimArrAt:0L, $
    dStimCurrent: 0.0d, $ ; stimulation amplitude
    ;pStimArr: OBJ_NEW(), $ ; stimulation array, always in ms resolution
    strStimType: '', $ ; stimulation type - point, linear, square, circle, all, ...
    strComments: '', $
    sDiff: 0, $ ; diffusion or not
    iBuffSize: 0L, $ ; buffer frames to record
    iBuffDim:0, $
    iAuxBuff:0, $
    iBuffInd:LONARR(10, 3), $ ; array index, celltypeindex, var index, type index,
    iBuffBinning:0L, $
    iBuffBinningInd:LONARR(3), $
    iArrDim:LONARR(2), $
    iFrameAt: 0L, $
    iNewNumCells:0L,$
    ssavepath: ' ', $
    fMiddleSD:0L, $ ; SD of Middle cell distribution
    fMiddleSize:0L, $ ; Middle cell block size
    fgNaFactor:0L, $ ; gNa Multiply factor of CF
    INHERITS ClsStcCellCltn $
  }
END

; initiation
PRO ClsMultiCellCuda::initObj
  COMMON CM_MULTI_CELL, CellTypeArr, StimArr, tempV, tempVRaw, Vind, BuffArr, iNeighborInd, couplingArr
  self.iDim = 2
  self.iDimXYZ = [25L, 25L, 25L]
  self.iDimXY = [125,125]
  self.dt = 0.001d
  self.dStimCurrent = -45.0d
  self.strStimLogic = 'S20.400.2'
  self.ssavepath = 'F:\'
  self->makeStimArr
  self.strStimType = 'all'
  self.iStimArrAt = 0
  self.iBuffSize = 2000L
  self.iBuffDim = 1 ; 1 variable
  self.iBuffInd = [0,0] ; voltage
  self.iAuxBuff = 1;
  self.dScanInt = 1.0
  self.fMiddleSD = 7L
  self.fMiddleSize = 6L
  self.fgNaFactor = 1.0 
  self->ClsStcCellCltn::initObj
END

; CELL INITIATION TO HERE

PRO ClsMultiCellCuda::initCMOnly2D_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 2
  ;ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125L]
  ;cxCoeff = [100.0d, 0.0d, 1.0d]
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  
  ; create cmcf array and neighbor ind
  createCMCFrandom2D, ratioCMCF, self.iDimXY, cxCoeff, cmcfArr,  iNeighborInd, couplingArr
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  help, CMind
  Print, iNCellCM
  iCMType = LONG(6) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(4)

  ;self->addCellTypeOnDevice, 'MahajanCM', LONG(iNCellCM)
  ;self.dStimCurrent = -120.0 ; for pandit, -60 amplitude  S1.20.5 : 5 ms stimulation
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self.dStimCurrent = -80.0 ; LuoRudyCM
  self->setCellLoc, 1L, CmInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCMType, LONG(iNCellCM))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 1L, LONG(CmInd))
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'PassiveCF', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCFType, LONG(iNCellCF))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 2L, LONG(CFInd))
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  ;self.dt = 0.0001d
  self.dt = 0.001d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice

  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
END

PRO ClsMultiCellCuda::initCMCF2D_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 2
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125L]
  cxCoeff = [50.0d, 5.0d, 5.0d]
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  ; res = CALL_EXTERNAL(dllName, 'IDL_alloc_ctn')
  ; create cmcf array and neighbor ind
  createCMCFrandom2D, ratioCMCF, self.iDimXY, cxCoeff, cmcfArr,  iNeighborInd, couplingArr
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  help, CMind
  Print, iNCellCM
  iCMType = LONG(6) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(4)

  ;self->addCellTypeOnDevice, 'MahajanCM', LONG(iNCellCM)
  ;self.dStimCurrent = -120.0 ; for pandit, -60 amplitude  S1.20.5 : 5 ms stimulation
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self.dStimCurrent = -80.0 ; LuoRudyCM
  self->setCellLoc, 1L, CmInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCMType, LONG(iNCellCM))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 1L, LONG(CmInd))
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'PassiveCF', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCFType, LONG(iNCellCF))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 2L, LONG(CFInd))
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  ;self.dt = 0.0001d
  self.dt = 0.01d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  ;res = CALL_EXTERNAL(dllName, 'IDL_alloc_diffusion', s[1])
  ;help, iNeighborInd, couplingArr
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  ; Print, min(iNeighborInd), max(iNeighborInd)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
END


PRO ClsMultiCellCuda::initCMCFmiddle2D_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 2
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125]
  cxCoeff = [100.0, 50.0, 50.0]
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  ; res = CALL_EXTERNAL(dllName, 'IDL_alloc_ctn')
  ; create cmcf array and neighbor ind
  ;createCMCFrandom2D, ratioCMCF, self.iDimXY, cxCoeff, cmcfArr,  iNeighborInd, couplingArr
  middlesize = 2L
  createCMCFmiddle2D, middlesize, self.iDimXY, cxCoeff, cmcfArr, iNeighborInd, couplingArr
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  help, CMind
  Print, iNCellCM
  iCMType = LONG(1) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(2)
  self->addCellTypeOnDevice, 'MahajanCM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCMType, LONG(iNCellCM))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 1L, LONG(CmInd))
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'MacCannellCF', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCFType, LONG(iNCellCF))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 2L, LONG(CFInd))
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  ;res = CALL_EXTERNAL(dllName, 'IDL_alloc_diffusion', s[1])
  ;help, iNeighborInd, couplingArr
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  ; Print, min(iNeighborInd), max(iNeighborInd)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
END

PRO ClsMultiCellCuda::initCMOnly2DLindear_onDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 2
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125]
  cxCoeff = [20.0, 20.0, 20.0]
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  middlesize = 5L
  createCMCFmiddle2DLinear, middlesize, self.iDimXY, cxCoeff, cmcfArr, iNeighborInd, couplingArr, mult=40.0
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  iCMType = LONG(1) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(2)
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  ;self.dt = 0.001d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
  ; adjust CM value
  ;; for CM
  CFind = WHERE(CellTypeArr EQ 2)
  nSize = N_ELEMENTS(CFind)
  CMind = WHERE(CellTypeArr EQ 1)
  cmnsize = N_ELEMENTS(CMind)
  vValue = DBLARR(nSize)
  iCell = 1L
  ;  ; vartypes: CellLoc, Rate, State, Const
  iVarType = 2L ; state
  iPar = 145L; I_CaL for LR1 ;index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  CMvValue = DBLARR(cmnSize)
  CMvValue[*] = .50 ; 120.0d ; g_xNa 12.00 default
  ;  vValue[*] = 70.0d ; Cm 6.3 default
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  iVarType = 2L ; state
  iPar = 144L; I_K1 for LR1 index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)

  iCell = 2
  vValue = DBLARR(nSize)
  vValue[*] = .50 ; 120.0d ; g_xNa 12.00 default
  iVarType = 2L ; state
  iPar = 145L; I_CaL for LR1 ;index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)
  iPar = 144L
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)
END


PRO ClsMultiCellCuda::initCMCFmiddle2DLinear_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 2
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125]
  cxCoeff = [500.0, 1.0, 1.0]
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  middlesize = 10L
  createCMCFmiddle2DLinearCMCF, middlesize, self.iDimXY, cxCoeff, cmcfArr, iNeighborInd, couplingArr, mult=500.0
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  help, CMind
  Print, iNCellCM
  iCMType = LONG(1) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(2)
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCMType, LONG(iNCellCM))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 1L, LONG(CmInd))
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'MacCannellCF', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCFType, LONG(iNCellCF))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 2L, LONG(CFInd))
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  self.dt = 0.001d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  ;res = CALL_EXTERNAL(dllName, 'IDL_alloc_diffusion', s[1])
  ;help, iNeighborInd, couplingArr
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  ; Print, min(iNeighborInd), max(iNeighborInd)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
  ; adjust CM value
  ;; for CM
  CFind = WHERE(CellTypeArr EQ 2)
  nSize = N_ELEMENTS(CFind)
  CMind = WHERE(CellTypeArr EQ 1)
  cmnsize = N_ELEMENTS(CMind)
  iCell = 1L
  ;  ; vartypes: CellLoc, Rate, State, Const
  iVarType = 2L ; state
  iPar = 145L; I_CaL for LR1 ;index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  CMvValue = DBLARR(cmnSize)
  CMvValue[*] = .50 ; 120.0d ; g_xNa 12.00 default
  ;  vValue[*] = 70.0d ; Cm 6.3 default
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  iVarType = 2L ; state
  CMvValue[*] = .10 ; 120.0d ; g_xNa 12.00 default
  iPar = 144L; I_K1 for LR1 index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)

END


PRO ClsMultiCellCuda::initCMCFmiddleTriangle2DLinear_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 2
  ;ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125]
  cxCoeff = [20.0, 20.0, 20.0]
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  ; res = CALL_EXTERNAL(dllName, 'IDL_alloc_ctn')
  ; create cmcf array and neighbor ind
  ;createCMCFrandom2D, ratioCMCF, self.iDimXY, cxCoeff, cmcfArr,  iNeighborInd, couplingArr
  middlesize = 30L
  createCMCFmiddleTriangle2DLinear, middlesize, self.iDimXY, cxCoeff, cmcfArr, iNeighborInd, couplingArr
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  help, CMind
  Print, iNCellCM
  iCMType = LONG(1) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(2)
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCMType, LONG(iNCellCM))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 1L, LONG(CmInd))
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'PassiveCF', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCFType, LONG(iNCellCF))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 2L, LONG(CFInd))
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info

  self.dt = 0.1d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)

  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  ;res = CALL_EXTERNAL(dllName, 'IDL_alloc_diffusion', s[1])
  ;help, iNeighborInd, couplingArr
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  ; Print, min(iNeighborInd), max(iNeighborInd)

  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
END


PRO ClsMultiCellCuda::initCMCF3D_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 3
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [125L, 125]
  self.iDimXYZ = [25, 25, 25]
  ;self.iDimXY = [216L, 216L]
  ;self.iDimXYZ = [36, 36, 36]
  cxCoeff = [12d, 6d, 6d]
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  res = CALL_EXTERNAL(dllName, 'IDL_alloc_ctn')
  ; create cmcf array and neighbor ind
  createCMCFrandom3D, ratioCMCF, self.iDimXYZ, cxCoeff, cmcfArr,  iNeighborInd, couplingArr
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  help, CMind
  Print, iNCellCM
  iCMType = LONG(6) ; 0,1,2 : Mahajan CM, Pandit rat CM, MacCannon CF model
  iCFType = LONG(4)
  ;self->addCellTypeOnDevice, 'MahajanCM', LONG(iNCellCM)
  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCMType, LONG(iNCellCM))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 1L, LONG(CmInd))
  iNCellCF = N_ELEMENTS(CFind)
  self->addCellTypeOnDevice, 'PassiveCF', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
  ;res = CALL_EXTERNAL(dllName, 'IDL_add_StcCell', iCFType, LONG(iNCellCF))
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_CellLoc', 2L, LONG(CFInd))
  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  self.dt = 0.01d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  ;res = CALL_EXTERNAL(dllName, 'IDL_alloc_diffusion', s[1])
  ;help, iNeighborInd, couplingArr
  ;res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  ; Print, min(iNeighborInd), max(iNeighborInd)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  self->makeBuffer
END

PRO ClsMultiCellCuda::initCMCF3DTube_OnDevice
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  self.iDim = 4 ; tube
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  self.iDimXY = [100L, 400L] ; (Z,X*Y)
  self.iDimXYZ = [100, 20, 20] ; define the size of tube (Z, X, Y)
  
  cxCoeff = [200.0d, 10.0d, 20.0d] ; define gap junction conductance(nS) (CM-CM, CM-CF, CF-CF)
  
  
  ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  res = CALL_EXTERNAL(dllName, 'IDL_alloc_ctn')
  ; create cmcf array and neighbor ind
  ;createCMCFrandom3D, ratioCMCF, self.iDimXYZ, cxCoeff, cmcfArr,  iNeighborInd, couplingArr

  self->getValue, 'MiddleFSD', fsd
  self->getValue, 'MiddleSize',fcentersize

  createCMCFmiddle3D, ratioCMCF, self.iDimXYZ, cxCoeff  , cmcfArr,  iNeighborInd, couplingArr, fsd, fcentersize

  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  ;  help, CMind
  Print, iNCellCM


  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd

  iNCellCF = N_ELEMENTS(CFind)
  ; self->addCellTypeOnDevice, 'PassiveCF', LONG(iNCellCF)
  ;self->addCellTypeOnDevice, 'PassiveIVCF', LONG(iNCellCF)
  self->addCellTypeOnDevice, 'MacCannellINaCF', LONG(iNCellCF)
  ;self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd


  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  self.dt = 0.01d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
  

  res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  Print, 'NCell' , self.N_CELL
  
  ; adjust CM value
  ;; for CM
  CFind = WHERE(CellTypeArr EQ 2)
  nSize = N_ELEMENTS(CFind)
  CMind = WHERE(CellTypeArr EQ 1)
  cmnsize = N_ELEMENTS(CMind)
  vValue = DBLARR(nSize)
  CMvValue = DBLARR(cmnSize)
  CFind = WHERE(CellTypeArr EQ 2)
  cfnsize = N_ELEMENTS(CFind)
  CFValue = DBLARR(cfnSize)
  iCell = 1L
  iVarType = 2L ; state
  iPar = 144L; I_K1 for LR1 index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
 ; CMvValue[*] = 0.5
  ;CFValue[*] = 0.5
 ; res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  ; res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', 2L, CFValue, iPar, iVarType)

  ; capacitance
  iCell = 1L
  iVarType = 2L ; state
  iPar = 7L;
  CMvValue[*] = 63.0
  CFValue[*] = 6.30
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', 2L, CFValue, iPar, iVarType)
  ;; gNa in LR
  ;;iVarType = 2L ; state
  ;;iPar = 12L; gNa, original = 0.16
  ;;CMvValue[*] = 0.16
  ;;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)

  ; for INa in MacCannell case

  iCell = 2L
  iVarType = 3L
  iPar = 7L
  CFValue[*] = (0.6) ;* 0.5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)

  iCell = 2L
  iVarType = 3L
  iPar = 8L
  CFValue[*] = (0.5) ;* 0.5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)

  ;; for background Na currents
  iCell = 2L
 iPar = 15L
  CFValue[*] = (0.0095)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)

  ; buffer
  self->makeBuffer
END

PRO ClsMultiCellCuda::updateCell3Dtube
  COMMON CM_MULTI_CELL
  dllname = self.dllname
  ratioCMCF = [1.0, 1.0] ; CM:CF = 1:1
  
  cxCoeff = [200.0d, 10.0d, 20.0d] ; define gap junction conductance(nS) (CM-CM, CM-CF, CF-CF)
  
    ; alloc collection obj in the cuda device
  memAval = 0L
  memTotal = 0L
  res = CALL_EXTERNAL(dllName, 'IDL_get_Cuda_Mem_info', memAval, memTotal)
  Print, 'Cuda memory: ', memTotal, memAval
  res = CALL_EXTERNAL(dllName, 'IDL_alloc_ctn')
  
  self->getValue, 'MiddleFSD', fsd
  self->getValue, 'MiddleSize',fcentersize

  createCMCFmiddle3D, ratioCMCF, self.iDimXYZ, cxCoeff  , cmcfArr,  iNeighborInd, couplingArr, fsd, fcentersize
  
  CMind = WHERE(cmcfArr EQ 1)
  CFind = WHERE(cmcfArr EQ 2)
  iNCellCM = N_ELEMENTS(CMind)
  
   self.N_CELL = 0L

  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCM)
  self->setCellLoc, 1L, CmInd

  iNCellCF = N_ELEMENTS(CFind)
  ; self->addCellTypeOnDevice, 'PassiveCF', LONG(iNCellCF)
  ;self->addCellTypeOnDevice, 'PassiveIVCF', LONG(iNCellCF)
  self->addCellTypeOnDevice, 'MacCannellINaCF', LONG(iNCellCF)
  ; self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd


  ctnInfo = LONARR(5)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_ctn', ctnInfo)
  PRINT, 'Ctn info: ', ctnInfo
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 1L, ctninfo)
  arr1 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst1 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 1 :', ctninfo ; checking the cell info
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_cell', 2L, ctninfo)
  arr2 = DBLARR(ctninfo[1],ctninfo[0])
  arrConst2 = DBLARR(ctninfo[2],ctninfo[0])
  Print, 'Cell 2 :', ctninfo ; checking the cell info
  self.dt = 0.01d
  res = CALL_EXTERNAL(dllName, 'IDL_set_dt', self.dt)
  s = SIZE(iNeighborInd)
  self->setDiffParm, iNeighborInd, couplingArr
  self->allocDiffDevice
 
  res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
  CellTypeArr = cmcfArr
  Print, 'NCell' , self.N_CELL
  ; adjust CM value
  ;; for CM
  CFind = WHERE(CellTypeArr EQ 2)
  nSize = N_ELEMENTS(CFind)
  CMind = WHERE(CellTypeArr EQ 1)
  cmnsize = N_ELEMENTS(CMind)
  vValue = DBLARR(nSize)
  CMvValue = DBLARR(cmnSize)
  CFind = WHERE(CellTypeArr EQ 2)
  cfnsize = N_ELEMENTS(CFind)
  CFValue = DBLARR(cfnSize)
  iCell = 1L
  iVarType = 2L ; state
  iPar = 144L; I_K1 for LR1 index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
 ; CMvValue[*] = 0.5
 ; CFValue[*] = 0.5
 ;  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
 ;  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', 2L, CFValue, iPar, iVarType)

  ; capacitance
  iCell = 1L
  iVarType = 2L ; state
  iPar = 7L;
  CMvValue[*] = 63.0
  CFValue[*] = 6.30
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', 2L, CFValue, iPar, iVarType)
  ;; gNa in LR
  ;;iVarType = 2L ; state
  ;;iPar = 12L; gNa, original = 0.16
  ;;CMvValue[*] = 0.16
  ;;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)

  ; for INa in MacCannell case

  iCell = 2L
  iVarType = 3L
  iPar = 7L
  CFValue[*] = (0.6) ;* 0.5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)

  iCell = 2L
  iVarType = 3L
  iPar = 8L
  CFValue[*] = (0.5) ;* 0.5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)

  ;; for background Na currents
  iCell = 2L
  iPar = 15L
  CFValue[*] = (0.0095)
 res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)

  ; buffer
  self->makeBuffer


END

PRO ClsMultiCellCuda::updateDiffusion
  self->multiplyDiffParm, 0.8
  self->updateDiffParmDevice
END

PRO ClsMultiCellCuda::updateINaCF, fFactor
  COMMON CM_MULTI_CELL
  ; for INa in MacCannell case
  ; adjust CM value
  ;; for CM

  
  CFind = WHERE(CellTypeArr EQ 2)
  nSize = N_ELEMENTS(CFind)
  CMind = WHERE(CellTypeArr EQ 1)
  cmnsize = N_ELEMENTS(CMind)
  vValue = DBLARR(nSize)
  CMvValue = DBLARR(cmnSize)
  CFind = WHERE(CellTypeArr EQ 2)
  cfnsize = N_ELEMENTS(CFind)
  CFValue = DBLARR(cfnSize)

  iCell = 1L
  iVarType = 2L ; state
  iPar = 144L; I_K1 for LR1 index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  CMvValue[*] = 1.0
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  ; g_Kp ; 26 (1.83e-4), g_Kmax; 23 (1.83e-4), g_K = 57
  ;iVarType = 2L ; state
  ;iPar = 57L
  ;CMvValue[*] = (2.82e-3)*4.0
  ;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  ;iPar = 23L
  ;CMvValue[*] = (2.82e-3)*4.0
  ;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  ;iPar = 26L
  ;CMvValue[*] = (1.83e-4)*4.0
  ;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)

  iCell = 2L
  iVarType = 3L
  iPar = 7L
  CFValue[*] = (0.6 * fFactor) ;* 0.5
  ;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)
  iCell = 2L
  iVarType = 3L
  iPar = 8L
  CFValue[*] = (0.5 * fFactor) ;* 0.5
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)
  ;; for background Na currents
  iCell = 2L
  iPar = 15L
  CFValue[*] = (0.0095 * 1.0)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)
  ;; capacitance
  iPar = 0L
  CFValue[*] = 6.3 * 1.0
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CFValue, iPar, iVarType)
END


PRO ClsMultiCellCuda::changeCMCFStateConst
  COMMON CM_MULTI_CELL
  ; enum CellMLVarType {CellLoc, Rate, State, Const};
  CFind = WHERE(CellTypeArr EQ 2)
  nSize = N_ELEMENTS(CFind)
  CMind = WHERE(CellTypeArr EQ 1)
  cmnsize = N_ELEMENTS(CMind)
  vValue = DBLARR(nSize)
  ;; for CM
  iCell = 1L
  ;  ; vartypes: CellLoc, Rate, State, Const
  iVarType = 2L ; state
  iPar = 145L; I_CaL for LR1 ;index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  CMvValue = DBLARR(cmnSize)
  CMvValue[*] = .50 ; 120.0d ; g_xNa 12.00 default
  ;  vValue[*] = 70.0d ; Cm 6.3 default
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)
  ;  ; vartypes: CellLoc, Rate, State, Const
  iVarType = 2L ; state
  iPar = 144L; I_K1 for LR1 index for MacCannonNa const Cm = 0, g_xNa = 7, g_Kv=5
  CMvValue = DBLARR(cmnSize)
  CMvValue[*] = 0.5 ; 120.0d ; g_xNa 12.00 default
  ;  vValue[*] = 70.0d ; Cm 6.3 default
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, CMvValue, iPar, iVarType)

  ; passive CF
  iCell = 2L
  iVarType = 2L ; state
  iPar = 1L; E_rev
  ;vValue[*] = -15.0d ; reversal potential
  ;; reversal potential random distribution from -60 to 0 mV
  fmean = -30.00
  ;rv = DOUBLE(RANDOMU(seed, nSize, /NORMAL) * 15.0 - fmean)
  ;rv <= (fmean + 15.0d) ; 0.0d
  ;rv >= (fmean - 15.0d) ; -60.0d
  vValue[*] = fmean ; rv[*]
  ;;vValue[*] = rv[*]
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)
  ;;iVarType = 2L ; state
  ;;iPar = 0L; initial Vm
  ;;; vValue[*] = -15.0d ; reversal potential
  ;;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)
  ;;;
  ;;; CF capacitance
  iVarType = 3L
  iPar = 0L
  vValue[*] = (50.0d)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)

  ;; CF g_pass
  iVarType = 3L
  iPar = 1L
  fmean = 4.000d
  ;rv = DOUBLE(RANDOMU(seed, nSize, /NORMAL) * 0.05 + fmean)
  ;rv <= (fmean + 0.05d) ;
  ;rv >= (fmean - 0.05d) ;
  vValue[*] = fmean ; rv[*] ; (.1d)
  res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)

  ; myocytes
  ;CMind = WHERE(CellTypeArr EQ 1)
  ;nSize = N_ELEMENTS(CMind)
  ;vValue = DBLARR(nSize)
  ;iCell = 1L
  ;iVarType = 3L
  ;iPar = 33L
  ;fmean = 0.36d / 2.0
  ;rv = DOUBLE(RANDOMU(seed, nSize, /NORMAL) * 0.05 + fmean)
  ;rv <= (fmean + 0.1d) ;
  ;rv >= (fmean - 0.1d) ;
  ;vValue[*] = rv[*] ; (0.36/2.0)
  ;res = CALL_EXTERNAL(self.dllName, 'IDL_set_CellValue', iCell, vValue, iPar, iVarType)

  ;self.dStimCurrent = 0.45 ; for LR1 model
END



PRO ClsMultiCellCuda::CLEANUP
  COMMON CM_MULTI_CELL
  Print, 'ClsMultiCellCuda freed'
  self->emptyCltn
  self->ClsStcCellCltn::CLEANUP
  PTR_FREE, self.pStimArr
  ;PTR_FREE, self.piNeighbors
  ;PTR_FREE, self.pdNeighborDiffCoeffs
END

PRO ClsMultiCellCuda::setValue, opt, val
  COMMON CM_MULTI_CELL
  CASE opt OF
    'Dim': BEGIN
      self.iDim = val
    END
    'Diffusion': self.sDiff = val
    'ScanInt': self.dScanInt = val
    'StimLogic': BEGIN
      self.strStimLogic = val
      self->makeStimArr
    END
    'SavePath':BEGIN
      self.ssavepath = val
    END
    'StimArrAt': BEGIN
      self.iStimArrAt = val
      self.iStimArrAt >= 0 ; self.iStimArrAt < 0
      self.iStimArrAt <= (N_ELEMENTS(*(self.pStimArr)) - 1); self.iStimArrAt
    END
    'FrameAt': self.iFrameat = val
    'BuffSize': BEGIN
      self.iBuffSize = val
      iNCells = self.N_Cell * LONG(self.iBuffDim)
      s = SIZE(BuffArr)
      xdim = iNCells+self.iAuxBuff
      ydim = self.iBuffSize
      IF ((xdim NE s[1]) OR (ydim NE s[2])) THEN BEGIN
        BuffArr = 0
        BuffArr = FLTARR(xdim, ydim)
      ENDIF
      ; BuffArr = FLTARR(iNCells+ self.iAuxBuff, self.iBuffSize )
    END
    'BuffDim': BEGIN
      self.iBuffDim = val
      iNCells = self.N_CELL * LONG(self.iBuffDim)
      s = SIZE(BuffArr)
      xdim = iNCells+self.iAuxBuff
      ydim = self.iBuffSize
      IF ((xdim NE s[1]) OR (ydim NE s[2])) THEN BEGIN
        BuffArr = 0
        BuffArr = FLTARR(xdim, ydim)
      ENDIF
    END
    'BuffInd': BEGIN
      self.iBuffInd[*,*] = 0
      s = SIZE(val)
      IF (s[0] EQ 1) THEN self.iBuffInd[0:s[1]-1,0] = val $
      ELSE IF (s[0] EQ 2) THEN BEGIN
        self.iBuffInd[0:s[1]-1, *] = val
      END
    END
    'BuffBinning': BEGIN
      self.iBuffBinning = val
      ;IF (val EQ 0) THEN BEGIN
      ;END
    END
    'BuffBinningInd': BEGIN
      s = SIZE(val)
      self.iBuffBinningInd[0:s[1]-1] = val
      self.iBuffBinning = 1
    END
    'CudaBlockSize': BEGIN
      IF (val GE 32) THEN res = CALL_EXTERNAL(self.dllname, 'IDL_set_Cuda_BlockSize', LONG(val))
    END
    'Comments': self.strComments = val
    'StimComments': self.strStimLogic = val

    'MiddleFSD' : self.fMiddleSD = val
    'MiddleSize' : self.fMiddleSize = val
    'gNaFactorCF': self.fgNaFactor = val
  ENDCASE
END

PRO ClsMultiCellCuda::getValue, opt, val
  COMMON CM_MULTI_CELL
  CASE opt OF
    'DiffObj': val = self.oDiff
    'CltnObj': val = self.oCltn
    'StimLogic': val = *(self.pStimArr)
    'StimLogicArr': val = *(self.pStimArr)
    'StrStimLogic': val = self.strStimLogic
    'ScanInt': val = self.dScanInt
    'BuffInd': val = self.iBuffInd
    'BuffDim': val = self.iBuffDim
    'BuffArr': val = BuffArr
    'BuffSize': val = SIZE(BuffArr)
    'AuxBuffSize': val = self.iAuxBuff
    'Dim': val = self.iDim
    'DimXY': val = self.iDimXY
    'ArrDim': val = self.iArrDim
    'StimLogic': val = *(self.pStimArr)
    'NumCells' : val = self.N_CELLS
    'CudaMemInfo': BEGIN
      a = 0L
      t = 0L
      res = call_external(self.dllName, 'IDL_get_Cuda_Mem_info', a, t)
      val = [a, t] ; available array size
    END
    'SavePath' : val = self.ssavepath
    'CellTypes': val = REFORM(CellTypeArr, self.iDimXY[0], self.iDimXY[1])
    'MiddleFSD' : val = self.fMiddleSD
    'MiddleSize' : val = self.fMiddleSize
     'gNaFactorCF': val = self.fgNaFactor

  ENDCASE
END

PRO ClsMultiCellCuda::getCellInfo, ind, val
  COMMON CM_MULTI_CELL
  ; iNeighborInd, couplingArr
  val = LONARR(4)
  val[0] = cellTypeArr[ind]
  iSize = SIZE(iNeighborInd)
  iNum = N_ELEMENTS(iNeighborInd)
  linearInd = REFORM(iNeighborInd,iSize[1], iNum/iSize[1])
  neigh = linearInd[*,ind]
  Print, neigh
  val[1] = N_ELEMENTS(neigh)
  cell = cellTypeArr[neigh]
  Print, cell
  num0 = (cell EQ 1)
  num1 = (cell EQ 2)
  val[2] = TOTAL(num0)
  val[3] = TOTAL(num1)
END


PRO ClsMultiCellCuda::makeBuffer
  COMMON CM_MULTI_CELL
  iNCells = self.N_CELL * LONG(self.iBuffDim)
  s = SIZE(BuffArr)
  xdim = iNCells+self.iAuxBuff
  ydim = self.iBuffSize
  IF (self.iBuffBinning GT 1) THEN BEGIN
  ENDIF
  IF ((xdim NE s[1]) OR (ydim NE s[2])) THEN BEGIN
    BuffArr = 0
    BuffArr = FLTARR(xdim, ydim)
  ENDIF
  tempV = DBLARR(self.N_CELL)
  IF (self.iDim EQ 2) THEN BEGIN
    stimArr = REFORM(DBLARR(self.N_CELL), self.iDimXY[0], self.iDimXY[1])
    tempVraw = tempV
  END
  IF (self.iDim EQ 3) THEN BEGIN
    tempVraw = REFORM(DBLARR(self.N_CELL), self.iDimXYZ[0], self.iDimXYZ[1], self.iDimXYZ[2])
    stimArr = REFORM(DBLARR(self.N_CELL), self.iDimXYZ[0], self.iDimXYZ[1], self.iDimXYZ[2])
    ind = REFORM(LINDGEN(self.N_CELL), self.iDimXYZ[0], self.iDimXYZ[1], self.iDimXYZ[2])
    vind = LONARR(self.iDimXY[0], self.iDimXY[1]) ; , self.iDimXYZ[2])
    istep = LONG(self.iDimXY[0]/self.iDimXYZ[2]) ; for array 3D to 2D tiling
    Print, self.iDimXY
    Print, self.iDimXYZ
    FOR z = 0L, self.iDimXYZ[2]-1 DO BEGIN
      xStart = (z MOD iStep) * self.iDimXYZ[0]
      xEnd = xStart + self.iDimXYZ[0]-1
      yStart = LONG(z/istep) * self.iDimXYZ[1]
      yEnd = yStart + self.iDimXYZ[1]-1
      ;Print, xstart, xend, ystart, yend
      vind[xStart:xEnd,yStart:yEnd]  = ind[*, *, z]
      ;ENDFOR
    ENDFOR
  END
  IF (self.iDim EQ 4) THEN BEGIN
    ;print, self.n_cell, self.idimxyz
    stimArr = REFORM(DBLARR(self.N_CELL), self.iDimXYZ[0], self.iDimXYZ[1], self.iDimXYZ[2])
    tempVraw = tempV
  ENDIF
END

PRO ClsMultiCellCuda::setBufferZero
  COMMON CM_MULTI_CELL
  BuffArr[*,*] = 0.0 ; = FLTARR(xdim, ydim)
  self.iFrameAt = 0L
END

PRO ClsMultiCellCuda::makeCouplingIndex2DCylinder ;, numX, numY, res
  COMMON CM_MULTI_CELL
  ;2D 4 side coupling + x direction sylinder
  numX = self.iArrDim[0]
  numY = self.iArrDim[1]
  self->makeCouplingIndex2D ;, numX, numY, res
  ; y axis index
  indY = LINDGEN(numY) * numX
  iNeighborIndex[0, indY] = indY+numX-1
  indYe = indY + numX -1
  iNeighborIndex[3, indYe] = indY
  ;res = iNeighborIndex
END


PRO ClsMultiCellCuda::copyNeighborIndexToDevice
  COMMON CM_MULTI_CELL
  help, iNeighborInd, couplingArr
  res = CALL_EXTERNAL(dllName, 'IDL_set_DiffParm', iNeighborInd, couplingArr)
  ; Print, min(iNeighborInd), max(iNeighborInd)
  res = CALL_EXTERNAL(dllName, 'IDL_get_info_diffusion', ctninfo)
  Print, 'diffusion : ', ctninfo
END

PRO ClsMultiCellCuda::copyCellsToHost
  strinfo = LONARR(5)
  res = CALL_EXTERNAL(self.dllname, 'IDL_get_info_ctn', strinfo)
  iNCell = strinfo[0]
  iNStcCell = strinfo[1]
  FOR I=0, iNStcCell-1 DO BEGIN
    self->copyCellFromDevice, I
  END
END


;PRO ClsMultiCellCuda::freeCellsOnCuda
;  res = CALL_EXTERNAL(self.dllname, 'IDL_free_Cells_on_cuda')
;  self.sCudaHasCells = 0
;END


PRO ClsMultiCellCuda::iterationNodiffusion, sSteps
  COMMON CM_MULTI_CELL
  res = CALL_EXTERNAL(self.dllName, 'IDL_simulate_NstepsNoDiffusion', sSteps)
END


PRO ClsMultiCellCuda::iteration, sSteps
  COMMON CM_MULTI_CELL
  res = CALL_EXTERNAL(self.dllName, 'IDL_simulate_Nsteps', sSteps)
END


PRO ClsMultiCellCuda::iteration1ms
  sStep = LONG(1.0 / self.dt)
  CASE self.sDiff OF
    0: self->iterationNoDiffusion, sStep
    1: self->iteration, sStep
  END
END

;PRO ClsMultiCellCuda::iteration1ms
;  sStep = LONG(1.0 / self.dt)
;  CASE self.sDiff OF
;     0: self->iterationNoDiffusion, sStep
;     1: self->iteration, sStep
;  END
;END

PRO ClsMultiCellCuda::iterationNms, Nms
  sStep = LONG(Nms / self.dt)
  CASE self.sDiff OF
    0: self->iterationNoDiffusion, sStep
    1: self->iteration, sStep
  END
END


PRO ClsMultiCellCuda::makeStimArr
  str = STRUPCASE(self.strStimLogic)
  str = STR_SEP(str, ' ')
  numElm = N_ELEMENTS(str)
  stimLogic = LONARR(numElm, 5)
  prod = [0]
  prevDur = 0
  FOR I=0, numElm - 1 DO BEGIN
    stimType = STRMID(str[I], 0, 1)	;Find the position of a substring. Find if it is R or S
    CASE stimType OF
      'S': BEGIN
        stimLogic[I,*] = getStimInfo(str[I])
        arrStim = makeStimArr(stimLogic[I,*])
        prod = [prod, arrStim[prevDur:*]]; Make array or add to it
        prevDur = stimLogic[I,3]
      END
      'R': BEGIN ; R200.250.2.30000
        stimLogic[I,*] = getRandomInfo(str[I])
        random_int = makeRandomStimLogic(stimLogic[I,*])
        random_arr = makeRandomArr(random_int, stimLogic[I,*])
        prod = [prod, random_arr[prevDur:*]]	; Make array, or add to it
        prevDur = stimLogic[I,3]
      END
      'O': BEGIN ; O20.350.25.2
        stimOsc = getOscillationInfo(str[I])
        arrStim = makeOscillationArr(stimOsc)
        prod = [prod, arrStim[prevDur:*]]; Make array or add to it
        prevDur = stimOsc[4]
      END
      'H': BEGIN ; H200.400.20.2
        stimLogicH = getHistStimInfo(str[I])
        arrStim = makeHistStimArr(stimLogicH)
        prod = [prod, arrStim[prevDur:*]]
        prevDur = stimLogic[4]
      END
      ELSE: ;Print, 'Unknown stimulation type'
    ENDCASE
  ENDFOR
  result = SHIFT(prod, -1) ; [prod, 0, 0] ; add zeros
  PTR_FREE, self.pStimArr
  self.pStimArr = PTR_NEW(result)
END


PRO ClsMultiCellCuda::writeNetCDF, filename, SAVEPATHINCLUDED=savepathincluded
  COMMON CM_MULTI_CELL
  ; read a file
  IF (KEYWORD_SET(savepathincluded)) THEN fsave = filename $
  ELSE fsave = self.ssavepath + filename
  ;self->getValue, 'SavePath', savepath
  ;fsave = savepath + filename

  IF (fileName EQ '') THEN RETURN
  id = NCDF_CREATE(fsave, /CLOBBER)
  NCDF_CONTROL, id, /FILL
  ; date        sTimeMonth:0, $
  d = BIN_DATE(systime())
  ; d = t[0:5] ; [t[0], h.sTimeMonth, h.sTimeDate, h.sTimeHour, h.sTimeMin, h.sTimeSec]
  NCDF_ATTPUT, id, /GLOBAL, 'DATE', d
  NCDF_ATTPUT, id, /GLOBAL, 'COMMENT', self.strStimLogic
  NCDF_ATTPUT, id, /GLOBAL, 'SCANINTERVAL', 1.0 ; 1 ms
  CASE self.iBuffDim OF
    1: BEGIN
      sDataType = 1 ; 1: vm only, 2: vm and Ca
      dim = [self.iArrDim[0], self.iArrDim[1], $
        self.iArrDim[0], self.iArrDim[1], $
        self.iNCells+ self.iAuxBuff, self.iNCells, self.iAuxBuff]
      iNCells = self.iNCells + self.iAuxBuff
    END
    2: BEGIN
      sDataType = 2 ; 1: vm only, 2: vm and Ca
      dim = [self.iArrDim[0], self.iArrDim[1]*2, $
        self.iArrDim[0], self.iArrDim[1], $
        self.iNCells*2 + self.iAuxBuff, self.iNCells, self.iAuxBuff]
      iNCells = self.iNCells * 2L + self.iAuxBuff
    END
  ENDCASE
  NCDF_ATTPUT, id, /GLOBAL, 'DATATYPE', sDataType ; 1: vm only, 2: vm and Ca
  NCDF_ATTPUT, id, /GLOBAL, 'DIMENSIONS', dim
  ; Put file in data mode:
  ; dimension
  ; h.iNumFrames = indRange[1] - indRange[0] + 1L
  xid = NCDF_DIMDEF(id, 'x', iNCells)    ; Make dimensions.
  yid = NCDF_DIMDEF(id, 'y', self.iBuffSize)
  ; Define variables:
  dataid = NCDF_VARDEF(id, 'Data', [xid, yid], /FLOAT)
  NCDF_CONTROL, id, /ENDEF
  ; put data:
  NCDF_VARPUT, id, dataid, BuffArr ; (*(self.pData))[*,indRange[0]:indRange[1]]
  NCDF_CLOSE, id ; Close the NetCDF file
END


PRO ClsMultiCellCuda::writeMultipleNetCDFs, filename, SAVEPATHINCLUDED=savepathincluded, FSAVE=FSAVE
  COMMON CM_MULTI_CELL
  ; read a file
  ; set saving path
  ; self->getValue, 'SavePath', savepath

  IF (fileName EQ '') THEN RETURN
  pname = FILE_DIRNAME(filename)
  bname = FILE_BASENAME(filename, '.nc')
  FOR I=0, self.iBuffDim-1 DO BEGIN
    IF (bname EQ '.') THEN fsave = ' ' ELSE fsave = bname
    IF (KEYWORD_SET(savepathincluded)) THEN fsave = pname + '\' + bname $
    ELSE fsave = self.ssavepath
    ;fsave = fsave  $
    ;      + '-var-' + STRING(self.iBuffInd[I,0], FORMAT='(I003)') $
    ;      + '-type-' + STRING(self.iBuffInd[I,1], FORMAT='(I003)') + '.nc'
    ;Print, fsave
    fsave = filename ; fsave + '.nc'
    id = NCDF_CREATE(fsave, /CLOBBER)
    NCDF_CONTROL, id, /FILL
    ; date        sTimeMonth:0, $
    d = BIN_DATE(systime())
    ; d = t[0:5] ; [t[0], h.sTimeMonth, h.sTimeDate, h.sTimeHour, h.sTimeMin, h.sTimeSec]
    NCDF_ATTPUT, id, /GLOBAL, 'DATE', d
    strComments = self.strComments + ' ' + self.strStimLogic
    Print, self.strComments, strComments
    NCDF_ATTPUT, id, /GLOBAL, 'COMMENT', strComments ; self.strStimLogic
    NCDF_ATTPUT, id, /GLOBAL, 'SCANINTERVAL', 1.0 ; 1 ms
    sDataType = 1 ; 1: vm only, 2: vm and Ca
    dim = [self.iDimXY[0], self.iDimXY[1], $
      self.iDimXY[0], self.iDimXY[1], $
      self.N_CELL + self.iAuxBuff, self.N_Cell, + self.iAuxBuff]
    NCDF_ATTPUT, id, /GLOBAL, 'DATATYPE', sDataType ; 1: vm only, 2: vm and Ca
    NCDF_ATTPUT, id, /GLOBAL, 'DIMENSIONS', dim
    ; Put file in data mode:
    ; dimension
    ; h.iNumFrames = indRange[1] - indRange[0] + 1L
    iNCells = self.N_CELL + self.iAuxBuff
    xid = NCDF_DIMDEF(id, 'x', iNCells)    ; Make dimensions.
    yid = NCDF_DIMDEF(id, 'y', self.iBuffSize)
    ; Define variables:
    dataid = NCDF_VARDEF(id, 'Data', [xid, yid], /FLOAT)
    NCDF_CONTROL, id, /ENDEF
    ; put data:
    iStart = self.N_CELL * I
    iEnd = iStart + self.N_CELL - 1 + self.iAuxBuff
    fsaveBuff = BuffArr[iStart:iEnd, *]
    fsaveBuff[self.N_CELL:*,*] = BuffArr[-self.iAuxBuff:*,*]
    NCDF_VARPUT, id, dataid, fsaveBuff ; (*(self.pData))[*,indRange[0]:indRange[1]]
    NCDF_CLOSE, id ; Close the NetCDF file
  ENDFOR
END

PRO ClsMultiCellCuda::adjustStimArr, st
  COMMON CM_MULTI_CELL
  ; StimArr[0:99, 0:99]= st

  ; stimulating only CM
  ;3d
  cfind = WHERE(cellTypeArr EQ 2)
  CASE (self.iDim) OF
    1: StimArr[*, *, *] = st
    2: StimArr[0:3,*] = st ; *0.5 ; 20.0
    3: StimArr[*, *, *] = st
    4: StimArr[0:3, 0:5, 0:5] = st
    ;4: StimArr[*, *, *] = st
  ENDCASE
  StimArr[cfInd] = 0.0
  ; Print, 'stimulation', st
  ;2d
  ; StimArr[100:200, 100:200]= st
END

PRO ClsMultiCellCuda::getVmFromDevice
  COMMON CM_MULTI_CELL
  ;Print, self.iDim
  CASE (self.iDim) OF
    2:BEGIN
      res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
    END
    3:BEGIN
      res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempVraw)
      ;tempV = TRANSPOSE(tempVraw, [0,1,2])
      ;help, tempV
      tempV = tempVraw[Vind]
      ;help, tempV
    END
    4:BEGIN
      res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempVraw)
      ;;tempV = TRANSPOSE(tempVraw, [0,1,2])
      ;;help, tempV
      ;tempV = tempVraw[Vind]
      ;;help, tempV
    END
    ELSE:  res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
  ENDCASE
END


PRO ClsMultiCellCuda::record, STIM=STIM
  COMMON CM_MULTI_CELL
  ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
  self->getVmFromDevice
  ;Print, self.iBuffDim
  CASE self.iBuffDim OF
    1: BEGIN
      BuffArr[0:self.N_CELL-1, self.iFrameAt] = tempV
    END
    2: BEGIN
      BuffArr[0:self.N_CELL-1, self.iFrameAt] = tempV
      ; get another variable
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_get_value', tempV, LONG(self.iBuffInd[1,0]), 1L) ;state
      loc = LONG(self.iBuffInd[1, 0])
      itype = LONG(self.iBuffInd[1,1])
      res = CALL_EXTERNAL(self.dllname, 'IDL_get_value', tempV, loc, itype)
      BuffArr[self.N_CELL:self.N_CELL*2-1, self.iFrameAt] = tempV
    END
    ELSE: BEGIN
      FOR I=0L, self.iBuffDim-1 DO BEGIN
        loc = LONG(self.iBuffInd[I, 0])
        itype = LONG(self.iBuffInd[I,1])
        res = CALL_EXTERNAL(self.dllname, 'IDL_get_value', tempV, loc, itype)
        iStart = self.N_CELL*I
        iEnd = iStart + self.N_CELL-1
        BuffArr[iStart:iEnd, self.iFrameAt] = tempV
      ENDFOR
    END
  ENDCASE
  IF (KEYWORD_SET(STIM)) THEN BuffArr[-1, self.iFrameAt] = STIM ; last index
  self.iFrameAt += 1L
  IF (self.iFrameAt GE self.iBuffSize) THEN self.iFrameAt = 0L
END


FUNCTION ClsMultiCellCuda::images
  COMMON CM_MULTI_CELL
  ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
  self->getVmFromDevice
  resImg = Buffarr[0:-2,0]
  ;self->reconstImg, tempV, oImg
  CASE self.iBuffDim OF
    1: BEGIN
      resImg[0:self.iNCells-1] = tempV
    END
    2: BEGIN ; not done yet
      resImg[0:self.iNCells-1] = tempV
      ; get another variable
      res = CALL_EXTERNAL(self.dllname, 'IDL_get_value', tempV, LONG(self.iBuffInd[1,0]), 1L) ;state
      resImg[self.iNCells:self.iNCells*2-1] = tempV
    END
    ELSE: BEGIN ; not done yet
      FOR I=0L, self.iBuffDim-1 DO BEGIN
        loc = LONG(self.iBuffInd[I, 0])
        itype = LONG(self.iBuffInd[I,1])
        res = CALL_EXTERNAL(self.dllname, 'IDL_get_value', tempV, loc, itype)
        iStart = self.iNCells*I
        iEnd = iStart + self.iNCells-1
        resImg[iStart:iEnd] = tempV
      ENDFOR
    END
  ENDCASE
  ;self.iFrames -= 1L
  RETURN, resImg
END


FUNCTION ClsMultiCellCuda::imagesFromBuffer
  COMMON CM_MULTI_CELL
  img = BuffArr[*, self.iFrameAt-1]
  RETURN, img
END

FUNCTION ClsMultiCellCuda::getTrace, iChn
  COMMON CM_MULTI_CELL
  res = BuffArr[iChn, *]
  RETURN, res
END


PRO ClsMultiCellCuda::runStimRecord
  COMMON CM_MULTI_CELL
  iNum = N_ELEMENTS( *(self.pStimArr) )
  iRecordStart = iNum - self.iBuffSize
  ;StimArr = REFORM(StimArr, self.iArrDim[0], self.iArrDim[1])
  FOR I = 0L, iRecordStart-1 DO BEGIN
    st = (*(self.pStimArr))[I] * self.dStimCurrent
    ;;StimArr[85:99, 0:15] = st
    ; StimArr[0:10, 0:10]= st
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    self->iteration1ms
    ;Print, i, iRecordStart
    IF (i MOD 10) EQ 0 THEN BEGIN
      Print, i, iRecordStart
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      self->getVmFromDevice
      tvscl, Reform(tempV, self.iArrDim[0],self.iArrDim[1])
    ENDIF

  ENDFOR
  self.iFrameAt = 0
  FOR J=iRecordStart, iNum-1 DO BEGIN
    st = (*(self.pStimArr))[J] * self.dStimCurrent
    ;;StimArr[85:99, 0:15] = st
    ;StimArr[0:10, 0:10]= st
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    self->iteration1ms
    self->record, STIM=st
    IF (J MOD 10) EQ 0 THEN BEGIN
      Print, J, iNum		; Print, J, iNum
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      self->getVmFromDevice
      tvscl, Reform(tempV, self.iArrDim[0],self.iArrDim[1])
    ENDIF
  ENDFOR
END

; iterate N ms without stimulation, no recording
PRO ClsMultiCellCuda::iterateN, iNum
  FOR I=0L, iNum-1 DO BEGIN
    self->iteration1ms
    self->record
    IF (I MOD 10) EQ 0 THEN Print, I, iNum
  ENDFOR
END

; iterate N ms with stimulation, no recording
PRO ClsMultiCellCuda::iterateStimN, iNum
  COMMON CM_MULTI_CELL
  iNumStim = N_ELEMENTS( *(self.pStimArr) )
  FOR I=0L, iNum-1 DO BEGIN
    st = (*(self.pStimArr))[self.iStimArrAt] * self.dStimCurrent
    ;StimArr[0:25, 0:25] = st ; currently just box stimulation
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    ; advance stimulation time
    self.iStimArrAt += 1L
    IF (self.iStimArrAt GE iNumStim) THEN self.iStimArrAt = 0
    self->iteration1ms
    self->record, STIM=st
    IF (I MOD 10) EQ 0 THEN Print, I, iNum
  ENDFOR
END

PRO ClsMultiCellCuda::iterateStimN_norecord, iNum
  COMMON CM_MULTI_CELL
  iNumStim = N_ELEMENTS( *(self.pStimArr) )
  FOR I=0L, iNum-1 DO BEGIN
    st = (*(self.pStimArr))[self.iStimArrAt] * self.dStimCurrent
    ;StimArr[0:25, 0:25] = st ; currently just box stimulation
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    ; advance stimulation time
    self.iStimArrAt += 1L
    IF (self.iStimArrAt GE iNumStim) THEN self.iStimArrAt = 0
    self->iteration1ms
    ;self->record, STIM=st
    ;IF (I MOD 10) EQ 0 THEN Print, I, iNum
  ENDFOR
END

; iterate buffer size and save the result on the frame buffer, no stimulation
PRO ClsMultiCellCuda::Acquire
  COMMON CM_MULTI_CELL
  self.iFrameAt = 0
  FOR I=0L, self.iBuffSize-1 DO BEGIN
    self->iteration1ms
    self->record
    IF (I MOD 10) EQ 0 THEN BEGIN
      Print, I, self.iBuffSize
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      self->getVmFromDevice
      tvscl, Reform(tempV, self.iArrDim[0],self.iArrDim[1])
    ENDIF

  ENDFOR
END

; iterate buffer size with stimulation and save the result on the frame buffer
PRO ClsMultiCellCuda::StimAcquire
  COMMON CM_MULTI_CELL
  self.iFrameAt = 0
  iNumStim = N_ELEMENTS( *(self.pStimArr) )
  FOR I=0L, self.iBuffSize-1 DO BEGIN
    st = (*(self.pStimArr))[self.iStimArrAt] * self.dStimCurrent
    ;StimArr[0:10, 0:10] = st
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    ; advance stimulation time
    self.iStimArrAt += 1L
    IF (self.iStimArrAt GE iNumStim) THEN self.iStimArrAT = 0
    self->iteration1ms
    self->record, STIM=st
    IF (I MOD 10) EQ 0 THEN BEGIN
      Print, I, self.iBuffSize
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      self->getVmFromDevice
      tvscl, Reform(tempV, self.iArrDim[0],self.iArrDim[1])
    ENDIF
  ENDFOR
END

; same as above but start recording after iDelay ms delay
PRO ClsMultiCellCuda::StimAcquireDelay, iDelay
  COMMON CM_MULTI_CELL
  ; DELAY
  self.iFrameAt = 0
  iNumStim = N_ELEMENTS( *(self.pStimArr) )
  FOR I=0L, iDelay-1 DO BEGIN
    st = (*(self.pStimArr))[self.iStimArrAt] * self.dStimCurrent
    ; StimArr[0:25, 0:25] = st
    ;StimArr[0:10, 0:10]= st
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    ; advance stimulation time
    self.iStimArrAt += 1L
    IF (self.iStimArrAt GE iNumStim) THEN self.iStimArrAT = 0
    self->iteration1ms
    IF (I MOD 10) EQ 0 THEN BEGIN
      Print, I, self.iBuffSize + iDelay
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      self->getVmFromDevice
      tvscl, Reform(tempV, self.iArrDim[0],self.iArrDim[1])
    ENDIF
  END
  ; RECORD
  FOR I=0L, self.iBuffSize-1 DO BEGIN
    st = (*(self.pStimArr))[self.iStimArrAt] * self.dStimCurrent
    ;StimArr[0:25, 0:25] = st
    ;StimArr[0:10, 0:10]= st
    self->adjustStimArr, st
    res = CALL_EXTERNAL(self.dllname, 'IDL_set_stim', StimArr)
    ; advance stimulation time
    self.iStimArrAt += 1L
    IF (self.iStimArrAt GE iNumStim) THEN self.iStimArrAT = 0
    self->iteration1ms
    self->record, STIM=st
    IF (I MOD 10) EQ 0 THEN BEGIN
      Print, I+iDelay, self.iBuffSize + iDelay
      ;res = CALL_EXTERNAL(self.dllname, 'IDL_getVm', tempV)
      self->getVmFromDevice
      tvscl, Reform(tempV, self.iArrDim[0],self.iArrDim[1])
    ENDIF
  ENDFOR
END





; read cells, ind
PRO readCellInfo, filename, cell_ind, cell_info, myostate, myoconst
  IF (fileName EQ '') THEN RETURN
  IF (~FILE_TEST(fileName, /REGULAR)) THEN RETURN
  finfo = FILE_INFO(fileName)
  ;self->freeDevice
  ;self->emptyCltn
  id = NCDF_OPEN(fileName, /NOWRITE)

  NCDF_ATTGET, id, /GLOBAL, 'N_STCCELL', n_stccell
  ;self.N_STCCELL = n_stccell
  NCDF_ATTGET, id, /GLOBAL, 'N_CELL', n_cell
  ;self.N_CELL = n_cell
  ;NCDF_ATTGET, id, /GLOBAL, 'NCELLS', ncells
  ;self.NCells = ncells
  NCDF_ATTGET, id, /GLOBAL, 'DIM', iDim
  ;self.iDim = iDim
  NCDF_ATTGET, id, /GLOBAL, 'DIMXY', iDimXY
  ;self.iDimXY = iDimXY
  NCDF_ATTGET, id, /GLOBAL, 'DIMXYZ', iDimXYZ
  ;self.iDimXYZ = iDimXYZ
  Print, 'n_stccell, n_cell, iDim, iDimXY, iDimXYZ'
  Print, n_stccell, n_cell, iDim, iDimXY, iDimXYZ
  ;cellTypeArr = LONARR(n_cell)

  IF cell_ind GE n_stccell THEN BEGIN
    PRINT, 'cell_ind is too big'
    RETURN
  ENDIF
  ;oObj = OBJ_NEW('ClsStcCell')
  strI = STRTRIM(string(cell_ind),1)
  cellvars = 'CellVar'+ strI
  NCDF_ATTGET, id, /GLOBAL,cellvars, var

  PRINT, var
  iCellType = var[0]
  N_STATE = var[1]
  N_CONST = var[2]
  N_CELL = var[3]
  cell_info = var
  statename = 'STATE'+strI
  NCDF_VARGET, id, statename, MyoState ; states
  ;oObj.ptrState = PTR_NEW(MyoState)
  constname = 'CONST'+ strI
  NCDF_VARGET, id, constname, MyoConst ; const
  Print, CellVars
  Print, MyoState[*,0]
  Print, MyoConst[*,0]

END


PRO createCouplingArray3D, cmcfArr, dimXYZ, cxCoeff, iNeighborInd, couplingArr
  iSize = SIZE(cmcfArr)
  linArr = REFORM(cmcfArr, N_ELEMENTS(cmcfArr))
  makeCouplingIndex3D, dimXYZ, iNeighborInd, indArr
  couplingArr = DOUBLE(iNeighborInd)
  couplingArr[*,*] = 0.0
  iGood = WHERE(iNeighborInd GE 0)
  print, 'n neighbor', n_elements(iGood)
  iGoodInd = iNeighborInd[iGood]
  iCenter = LONG(iGood / 6L)
  iType = linArr[iCenter] + linArr[iGoodInd]
  ; CMCM = 1 + 1 = 2
  ; CMCF = 1 + 2 = 3
  ; CFCF = 2 + 2 = 4
  iCMCMInd = WHERE(iType EQ 2)
  print, 'CMCM', n_elements(iCMCMInd)
  couplingArr[iGood[iCMCMInd]] = cxCoeff[0]
  iCMCFInd = WHERE(iType EQ 3)
  print, 'CMCF', n_elements(iCMCFInd)
  couplingArr[iGood[iCMCFInd]] = cxCoeff[1]
  iCFCFInd = WHERE(iType EQ 4)
  print, 'CFCF', n_elements(iCFCFInd)
  couplingArr[iGood[iCFCFInd]] = cxCoeff[2]
END


PRO makeCouplingIndex3D, dimXYZ, iNeighborInd, indArr
  iNeighborSize = 6L
  numX = dimXYZ[0]
  numY = dimXYZ[1]
  numZ = dimXYZ[2]
  iNeighborInd = LONARR(iNeighborSize, numX, numY, numZ) ;, iNeighborSize)
  d = [[-1,0, 0], [0,1, 0], [0,-1, 0], [1,0, 0], [0,0, -1], [0,0,1]] ; 6 sides
  ; stupid C code style
  ; index array with edges
  indArr = LONARR(numX+2, numY+2, numZ+2)
  indArr[*,*,*] = -1
  ind = LINDGEN(numX, numY, numZ)
  ;indAdj = TRANSPOSE(ind, [0,2,1])
  indArr[1:numX, 1:numY, 1:numZ] = ind
  FOR Z=1L, numZ DO BEGIN
    FOR Y=1L, numY DO BEGIN
      FOR X=1L, numX DO BEGIN
        iNeighborInd[0,X-1,Y-1,Z-1] = indArr[X-1, Y, Z]
        iNeighborInd[1,X-1,Y-1,Z-1] = indArr[X, Y+1, Z]
        iNeighborInd[2,X-1,Y-1,Z-1] = indArr[X, Y-1, Z]
        iNeighborInd[3,X-1,Y-1,Z-1] = indArr[X+1, Y, Z]
        iNeighborInd[4,X-1,Y-1,Z-1] = indArr[X, Y, Z-1]
        iNeighborInd[5,X-1,Y-1,Z-1] = indArr[X, Y, Z+1]
        ; z need to be firt
      ENDFOR ; X
    ENDFOR ; Y
  ENDFOR ; Z
  ; neighbor index
  ;     3
  ;  1  C  4
  ;     2
  ; z -1, +1
END

PRO createCMCFrandom3D, ratioCMCF, dimxyz, cxCoeff, cmcfArr,  iNeighborInd, couplingArr
  iNum = dimXYZ[0] * dimXYZ[1] * dimXYZ[2]
  ; cxCoeff = [10.0, 3.0, 1.0]
  tempArr = RANDOMU(seed, iNum, /UNIFORM)
  indSorted = SORT(tempArr)
  fThr = LONG(FLOAT(ratioCMCF[0]) / (FLOAT(ratioCMCF[0]) + FLOAT(ratioCMCF[1])) * iNum)
  cmcfArr = LONARR(iNum)
  cmcfArr[*] = 2
  cmcfArr[indSorted[0:fThr]]=1
  cmcfArr = REFORM(cmcfArr, dimXYZ[0], dimXYZ[1], dimXYZ[2])
  iNumCM = N_ELEMENTS(WHERE(cmcfarr EQ 1L))
  iNumCF = N_ELEMENTS(WHERE(cmcfarr EQ 2L))
  info = [iNumCM, iNumCF]
  Print, info
  createCouplingArray3D, cmcfArr, dimXYZ, cxCoeff, iNeighborInd, couplingArr
END

FUNCTION createDensityCMCFmiddle3DPlot, dimxyz
  iNum = dimXYZ[0] * dimXYZ[1] * dimXYZ[2]
  iNumLayer = dimXYZ[1] * dimXYZ[2]
  ;CASE option OF
  ;  'Rectangular':
  ;  '
  ;ENDCASE
  iMax = 15
  fSD1 = 8.0
  fSD2 = 8.0
  re = DBLARR(dimXYZ[0], iMax)
  FOR I=1, iMax DO BEGIN
    fCenterSize = 7
    iStartLoc = LONG(dimXYZ[0]/2 - fCenterSize/2) - 1
    ;iStartLoc = 49
    ;print, istartloc
    fSD1 = FLOAT(I)
    fSD2 = FLOAT(I)
    iStartLoc2 = iStartLoc + fCenterSize
    iHalf = LONG(dimXYZ[0]/2)
    fG1 = GAUSSIAN_FUNCTION(fSD1, WIDTH=dimXYZ[0])
    fG2 = GAUSSIAN_FUNCTION(fSD2, WIDTH=dimXYZ[0])
    fg = REPLICATE(1.0, dimXYZ[0])
    ;help, fg2
    ;PLOT, FG2
    fg[0:iStartLoc] = fg1[(iHalf-iStartLoc-1):iHalf-1]
    Print, i, istartLoc, iStartLoc2, iHalf
    iSize = dimXYZ[0] - iStartLoc2
    fg[iStartLoc2:*] = fg2[iHalf:iHalf+iSize-1]
    re[*,I-1] = fg
    if (i EQ 1) then BEGIN
      p = plot(fg, YSTY=3, YRANGE=[0.0, 1.1])
    endif else BEGIN
      p = plot(fg, /OVERPLOT)
    endelse
  ENDFOR
  RETURN, re
END

PRO createCMCFmiddle3D, ratioCMCF, dimxyz, cxCoeff, cmcfArr,  iNeighborInd, couplingArr, fsd, fCenterSize;, option=option
  iNum = dimXYZ[0] * dimXYZ[1] * dimXYZ[2]
  iNumLayer = dimXYZ[1] * dimXYZ[2]
  ;CASE option OF
  ;  'Rectangular':
  ;  '
  ;ENDCASE
  fSD1 = fsd ; 12; 27.0
  fSD2 = fsd ; 12; 27.0
  ;fCenterSize = 5
  iStartLoc = LONG(dimXYZ[0]/2 - fCenterSize/2)-1
  ;iLeftSize = 41
  print, istartloc
  iStartLoc2 = iStartLoc + fCenterSize
  iHalf = LONG(dimXYZ[0]/2)
  fG1 = GAUSSIAN_FUNCTION(fSD1, WIDTH=dimXYZ[0])
  fG2 = GAUSSIAN_FUNCTION(fSD2, WIDTH=dimXYZ[0])
  fg = REPLICATE(1.0, dimXYZ[0])
  ;help, fg2
  ;PLOT, FG2
  fg[0:iStartLoc] = fg1[(iHalf-iStartLoc-1):iHalf-1]
  iSize = dimXYZ[0] - iStartLoc2
  fg[iStartLoc2:*] = fg2[iHalf:iHalf+iSize-1]
  ;fg *= 0.95
  ; 5% each case
  ;fg *= 0.90
  ;fg += 0.05
  numfG = LONG(fG*iNumLayer)
  ;Print, numfG
  iLayer = LONARR(dimXYZ[1], dimXYZ[2])
  cmcfArr = LONARR(dimXYZ[0], dimXYZ[1], dimXYZ[2])
  cmcfArr[*,*,*] = 1
  FOR I=0, dimXYZ[0] - 1 DO BEGIN
    IF (numfG[I] GE 1) THEN BEGIN
      iLayer[*,*] = 1
      tempArr = RANDOMU(seed, iNumLayer, /UNIFORM)
      ;help, indsorted
      ;numfg[I]
      ;Print, I, numfG[I]
      indSorted = SORT(tempArr)
      iLayer[indSorted[0:numfG[I]-1]]=2
      cmcfArr[I,*,*] = iLayer
    ENDIF
  ENDFOR
  iNumCM = N_ELEMENTS(WHERE(cmcfarr EQ 1L))
  iNumCF = N_ELEMENTS(WHERE(cmcfarr EQ 2L))
  info = [iNumCM, iNumCF]
  CreateCouplingArray3D, cmcfArr, dimXYZ, cxCoeff, iNeighborInd, couplingArr
END

PRO ClsMultiCellCuda::changeCellType, iNCellCF

  self->addCellTypeOnDevice, 'LuoRudy1CM', LONG(iNCellCF)
  self->setCellLoc, 2L, CFInd
END
