
/* Modified for multicellcular version 
   using struct
   can be called from an iterpreter languages
   use cuda approach to accelerate calculation
   GNU public license
   Copyright to Bum-Rak Choi
   Oct 3, 2014 */

 
#ifndef StcCell_H
#define StcCell_H

#ifdef __cplusplus 
	extern "C" {
#endif

#define iCudaBlockSize 32

/* List of variables and paramaters */ 
#define PI 3.1415926535897932846264
typedef double TypeCellML;

typedef TypeCellML TypeAlgebraic;
typedef TypeCellML TypeRate;
typedef TypeCellML TypeState;
typedef TypeCellML TypeConstant;
typedef TypeCellML TypeDiffusion;

enum CellMLVarType {CellLoc, Rate, State, Const};
enum CellType {CELL_MAHAJAN_CM, CELL_PANDIT_RATCM, CELL_MACCANNELL_CF, CELL_MACCANNELLINA_CF,CELL_PASSIVE_CF, CELL_LUORUDY_CM, CELL_LUORUDY1_CM, CELL_PASSIVEIV_CF};

struct StcCell_ {
	int iCellType;
	int N_ALGEBRAIC;
	int N_RATE;
	int N_STATE_ORG;
	int N_STATE; /* additional two for diffusion and stimulation */
	int N_CONST;
	int LOC_STIM;
	int LOC_DIFF;
	int LOC_Vm;
	int N_CELL; /* number of cells */
	int isAllocated; /* for memory clear */
	int iMemFlagCellLoc;
	int iMemFlagAlgebraic;
	int iMemFlagRate;
	int iMemFlagState;
	int iMemFlagConst;
	int iMemFlagTemp;
	//int *d_CellFlag; /* cuda cell flag */
	int *d_CellLoc; /* cell locaiton */
	TypeCellML *d_Algebraic; /* not used */
	TypeCellML *d_Rate; /* memory allocation on cuda device */ 
	TypeCellML *d_State;
	TypeCellML *d_Const;
	TypeCellML *d_temp;
	
	void (*fpAllocMemoryOnCuda)(struct StcCell_ *);
	void (*fpInitCellPar)(struct StcCell_ *);
	void (*fpInitCellParHost)(TypeCellML *, TypeCellML *, int);
	void (*fpIteration)(struct StcCell_ *, TypeCellML);
	void (*fpGetVm)(struct StcCell_ *, TypeCellML);

};

typedef struct StcCell_ StcCell;

int  cell_alloc_StcCellOnCuda(StcCell *);
void cell_free_StcCellOnCuda(StcCell *);

// host device copy
void cell_copy_VarToCuda(StcCell *, TypeCellML *, enum CellMLVarType );
void cell_copy_VarFromCuda(StcCell *, TypeCellML *, enum CellMLVarType );
// host device copy single var
void cell_set_CellValue(StcCell *, TypeCellML *, int , enum CellMLVarType );
void cell_get_CellValue(StcCell *, TypeCellML *, int , enum CellMLVarType );
// device device copy single var
void cell_copy_CellValueDevice(StcCell *, TypeCellML *, int , enum CellMLVarType , int );
// sorted set-get function for irregular array
void cell_get_CellValueSorted(StcCell *, TypeCellML *, int, enum CellMLVarType);
void cell_set_CellValueSorted(StcCell *, TypeCellML *, int, enum CellMLVarType);

// electrotonic coupling
void cell_get_VmSorted(StcCell *, TypeCellML *);
void cell_set_VmDiffSorted(StcCell *, TypeCellML *);

#ifdef __cplusplus
}
#endif

#endif

