#include <stdlib.h>
#include <string.h>
#include <cuda.h>
#include <stdio.h>
#include "StcCell_Container.h"


StcCellContainer *ctn_prototype(void) 
{
	StcCellContainer *ptr;
	ptr = (StcCellContainer *)malloc(sizeof(StcCellContainer));
	ptr->iNCell = 0;
	ptr->iNStcCell = 0;
	ptr->iMemFlag_temp = 0;
	return ptr;
}


void ctn_freeStcCellContainer(StcCellContainer *pCont)
{
	int i;
	StcCell *pStc;
	i = pCont->iNStcCell;
	while (i>0) {
		pStc = pCont->pCell[i];
		cell_free_StcCellOnCuda(pStc);
		free(pStc);
		i--;
	}
	if (pCont->iMemFlag_temp) cudaFree(pStc->d_temp);
	pCont->iMemFlag_temp = 0;
	free(pCont);
}


int ctn_addCell(StcCellContainer *pCont, StcCell *pStc)
{
	cudaError_t e;

	pCont->iNStcCell++;
	pCont->pCell[pCont->iNStcCell] = pStc;
	pCont->iNCell += pStc->N_CELL;
	if (pCont->iMemFlag_temp) cudaFree(pCont->d_temp);
	e = cudaMalloc((void **)&(pCont->d_temp), pCont->iNCell * sizeof(TypeCellML));
	pCont->iMemFlag_temp = 1;
	return e;
}


void ctn_iterate(StcCellContainer *pCont, TypeCellML dt)
{
	int i;
	StcCell *pStc;
	for (i=1;i<=pCont->iNStcCell;i++) {
		pStc = pCont->pCell[i];
		(*(pStc->fpIteration))(pStc, dt);
	}
}


void ctn_getValueSorted_device(StcCellContainer *pCtn, int *iLoc, enum CellMLVarType t)
{
	int i;
	StcCell *pCell;
	for(i=1;i<=pCtn->iNStcCell;i++) {
		pCell = pCtn->pCell[i];
		cell_get_CellValueSorted(pCell, pCtn->d_temp, iLoc[i], t);
	}
}

void ctn_getCtnValueToHost(StcCellContainer *pCtn, TypeCellML *ptr)
{
	if (pCtn->iMemFlag_temp) {
		cudaMemcpy(ptr, pCtn->d_temp, pCtn->iNCell * sizeof(TypeCellML), cudaMemcpyDeviceToHost);
	}
}

void ctn_getVm(StcCellContainer *pCtn, TypeCellML *pVal)
{
	StcCell *pCell;
	int i;
	for(i=1;i<=pCtn->iNStcCell;i++) {
		pCell = pCtn->pCell[i];
		cell_get_VmSorted(pCell, pCtn->d_temp);
	}
	ctn_getCtnValueToHost(pCtn, pVal);
}


void ctn_setVm(StcCellContainer *pCtn, TypeCellML *pVal)
{
	StcCell *pCell;
	int i;
	cudaMemcpy(pCtn->d_temp, pVal, pCtn->iNCell * sizeof(TypeCellML), cudaMemcpyHostToDevice);
	for(i=1;i<=pCtn->iNStcCell;i++) {
		pCell = pCtn->pCell[i];
		cell_set_CellValueSorted(pCell, pCtn->d_temp, pCell->LOC_Vm, State);
	}
}

void ctn_setStim(StcCellContainer *pCtn)
{
	StcCell *pCell;
	int i;
	for(i=1;i<=pCtn->iNStcCell;i++) {
		pCell = pCtn->pCell[i];
		cell_set_CellValueSorted(pCell, pCtn->d_temp, pCell->LOC_STIM, State);
	}
}

void ctn_setTemp(StcCellContainer *pCtn, TypeCellML *pVal)
{
	cudaMemcpy(pCtn->d_temp, pVal, pCtn->iNCell * sizeof(TypeCellML), cudaMemcpyHostToDevice);
}


void ctn_setValue_SingleCellType(StcCellContainer *pCtn, TypeCellML *ptr, int iLoc, CellMLVarType iType, int iCell)
{
	StcCell *pCell;
	pCell = pCtn->pCell[iCell];
	cell_set_CellValue(pCell, ptr, iLoc,  iType);
}



