#include "StcCell_Diffusion.h"

// prototype
StcCellDiffusion *diff_prototype_StcCellDiffusion(StcCellContainer *pContainer, int iMaxCoupling)
{
	StcCellDiffusion *pStc;
	pStc = (StcCellDiffusion *)malloc(sizeof(StcCellDiffusion));
	pStc->pCtn = pContainer;
	pStc->iNCell = pContainer->iNCell;
	pStc->iNCoupling = iMaxCoupling;
	diff_alloc_StcCellDiffusion(pStc);
	return pStc;
}


// memory allocation
void diff_alloc_StcCellDiffusion(StcCellDiffusion *pStc)
{
	cudaError_t e;
	int iByteSize;
	
	iByteSize = pStc->iNCell * sizeof(int) * pStc->iNCoupling;
	e = cudaMalloc( (void **)&(pStc->d_NeighborInd), iByteSize);
	iByteSize = pStc->iNCell * sizeof(TypeCellML) * pStc->iNCoupling;
	e = cudaMalloc( (void **)&(pStc->d_DiffCoeff), iByteSize);
	iByteSize = pStc->iNCell * sizeof(TypeCellML);
	e = cudaMalloc( (void **)&(pStc->d_diff), iByteSize);
	iByteSize = pStc->iNCell * sizeof(TypeCellML);
	e = cudaMalloc( (void **)&(pStc->d_Vm), iByteSize);
	pStc->isAllocated = 1;
}

void diff_free_StcCellDiffusion(StcCellDiffusion *pStc)
{
	if (pStc->isAllocated) {
		cudaFree(pStc->d_NeighborInd);
		cudaFree(pStc->d_DiffCoeff);
		cudaFree(pStc->d_diff);
		cudaFree(pStc->d_Vm);
		pStc->isAllocated = 0;
	}
}


void diff_set_DiffParm(StcCellDiffusion *pStc, int *h_NeighborInd, TypeCellML *h_DiffCoeff)
{
	int iByteSize;
	
	// diffusion index
	iByteSize = sizeof(int) * pStc->iNCell * pStc->iNCoupling;
	cudaMemcpy(pStc->d_NeighborInd, h_NeighborInd, iByteSize, cudaMemcpyHostToDevice);
	iByteSize = sizeof(TypeCellML) * pStc->iNCell * pStc->iNCoupling;
	cudaMemcpy(pStc->d_DiffCoeff, h_DiffCoeff, iByteSize, cudaMemcpyHostToDevice);
}


void diff_get_VmFromCell(StcCellDiffusion *pStcDiff)
{
	int i;
	StcCellContainer *pCtn;
	StcCell *pCell;

	pCtn = pStcDiff->pCtn;
	for (i=1;i<=pCtn->iNStcCell;i++) {
		pCell = pCtn->pCell[i];
		cell_get_VmSorted(pCell, pStcDiff->d_Vm); 
	}		
}


void diff_put_DiffOnCell(StcCellDiffusion *pStc)
{
	int i;
	StcCellContainer *pCtn;
	StcCell *pCell;

	pCtn = pStc->pCtn;
	
	for (i=1;i<= pCtn->iNStcCell;i++) {
		pCell = pCtn->pCell[i];
		cell_set_VmDiffSorted(pCell, pStc->d_diff);
	}		
}


__global__ void kern_calcGeneralDiffusion(TypeCellML *d_Vm, TypeCellML *d_diff, int *d_iNeighbor, TypeCellML *d_DiffCoeff, int iNCoupling, int iNCell){		
	int i, iLoc, iNe; //, iNeFlag;
	TypeCellML lapval;
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iNCell) {
		lapval = 0.0;
		for (i=0;i<iNCoupling;i++) {
			iNe = idx*iNCoupling + i;
			iLoc = d_iNeighbor[iNe];
			if (iLoc >=0) {
				lapval += ( (d_Vm[iLoc]-d_Vm[idx] ) * d_DiffCoeff[iNe] );
			}
		}
		d_diff[idx] = lapval; 
	}
}



void diff_calc_Diffusion(StcCellDiffusion *pStc) {
	int iBlockDim;
	iBlockDim = (pStc->iNCell / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStc->iNCell) iBlockDim ++;
	diff_get_VmFromCell(pStc);
	kern_calcGeneralDiffusion <<< iBlockDim, iCudaBlockSize>>> 
		(pStc->d_Vm, pStc->d_diff, pStc->d_NeighborInd, pStc->d_DiffCoeff, 
			pStc->iNCoupling, pStc->iNCell) ;
	diff_put_DiffOnCell(pStc);
}


