/* IDL iterpreter wrapper */

#include <stdlib.h>
#include "StcCell_Cuda_Interface.h"
 

int IDL_set_Cuda_Device(int argc, void* argv [])
{
	int iDev, res;
	iDev = *(int *) argv[0];
	res = cuda_set_DeviceId (iDev);
	return res;
}

int IDL_set_Cuda_BlockSize(int argc, void* argv [])
{
	int iSize;
	iSize = *(int *) argv[0];
	cuda_set_BlockSize(iSize);
	return 1;
}

int IDL_get_Cuda_Mem_info(int argc, void* argv [])
{
	size_t available, total;
	int	*a, *t;
	a = (int *)argv[0];
	t = (int *)argv[1];
	cuda_meminfo(&available, &total);
	*a = (int) (available / sizeof(TypeCellML));
	*t = (int) (total / sizeof(TypeCellML));	
	return 1;
}


// iNCell, iNStcCell, iMemFlag_temp
int IDL_get_info_ctn(int argc, void* argv[])
{
	int *pVal;
	pVal = (int *)argv[0];
	cuda_get_info_ctn(pVal);
	return 1;
}


// N_CELL, N_STATE, N_CONST
int IDL_get_info_cell(int argc, void* argv[])
{
	int iCell, *pVal;
	iCell = *(int *)argv[0];
	pVal = (int *)argv[1];
	cuda_get_info_cell(iCell, pVal);
	return 1;
}

// iNCell, iNCoupling, isAllocated
int IDL_get_info_diffusion(int argc, void *argv[])
{
	int *pVal;
	pVal = (int *)argv[0];
	cuda_get_info_diffusion(pVal);
	return 1;
}


int IDL_alloc_ctn(int argc, void *argv[])
{
	cuda_alloc_ctn();
	return 1;
}

int IDL_free_ctn(int argc, void *argv[])
{
	cuda_free_ctn();
	return 1;
}

int IDL_alloc_diffusion(int argc, void *argv[])
{
	int iMaxCoupling;
	iMaxCoupling = *(int *)argv[0];
	cuda_alloc_diffusion(iMaxCoupling);
	return 1;
}

int IDL_free_diffusion(int argc, void *argv[])
{
	cuda_free_diffusion();
	return 1;
}

int IDL_add_StcCell(int argc, void* argv[])
{
	//enum CellType c;
	int  i, iNCell;
	StcCell *pCell;

	i = *(int *)argv[0];
	//c = (CellType) i;
	iNCell = *(int *)argv[1];
	cuda_add_cell(i, iNCell);
	return 1;	
}

int IDL_set_CellLoc(int argc, void *argv[])
{
	int iCell;
	int *pLoc;
	iCell = *(int *)argv[0];
	pLoc = (int *)argv[1];
	cuda_set_CellLoc(iCell, pLoc);
	return 1;
}

int IDL_get_CellLoc(int argc, void *argv[])
{
	int iCell;
	int *pLoc;
	iCell = *(int *)argv[0];
	pLoc = (int *)argv[1];
	cuda_get_CellLoc(iCell, pLoc);
	return 1;
}



int IDL_set_dt(int argc, void *argv[])
{
	TypeCellML dt;
	dt = *(TypeCellML *)argv[0];
	cuda_set_dt(dt);
	return 1;
}


int IDL_copy_state_ToDevice(int argc, void *argv[])
{
	int iCell;
	TypeCellML *pArr;
	iCell = *(int *)argv[0];
	pArr = (TypeCellML *)argv[1];
	cuda_set_Cell_State(iCell, pArr);
	return 1;
}


int IDL_copy_const_ToDevice(int argc, void *argv[])
{
	int iCell;
	TypeCellML *pArr;
	iCell = *(int *)argv[0];
	pArr = (TypeCellML *)argv[1];
	cuda_set_Cell_Const(iCell, pArr);
	return 1;
}


int IDL_copy_state_ToHost(int argc, void *argv[])
{
	int iCell;
	TypeCellML *pArr;
	iCell = *(int *)argv[0];
	pArr = (TypeCellML *)argv[1];
	cuda_get_Cell_State(iCell, pArr);
	return 1;
}


int IDL_copy_const_ToHost(int argc, void *argv[])
{
	int iCell;
	TypeCellML *pArr;
	iCell = *(int *)argv[0];
	pArr = (TypeCellML *)argv[1];
	cuda_get_Cell_Const(iCell, pArr);
	return 1;
}

int IDL_set_CellValue(int argc, void *argv[])
{
	int iCell, iPar;
	int iVarType;
	TypeCellML *pVal;

	iCell = *(int *)argv[0];
	pVal = (TypeCellML *)argv[1];
	iPar = *(int *)argv[2];
	iVarType = *(int *)argv[3];
	cuda_set_CellValue(iCell, pVal, iPar, iVarType);
	return 1;
}


int IDL_getValueSorted(int argc, void *argv[])
{
	TypeCellML *pArr;
	int *iIndCell;
	int iVarType;

	pArr = (TypeCellML *)argv[0];
	iIndCell = (int *)argv[1];
	iVarType = *(int *)argv[2];
	cuda_get_ValueSorted(pArr, iIndCell, iVarType);
	return 1;
}

int IDL_getVm(int argc, void *argv[])
{
	TypeCellML *pArr;
	pArr = (TypeCellML *) argv[0];
	cuda_get_Vm(pArr);
	return 1;
}


int IDL_set_stim(int argc, void *argv[])
{
	TypeCellML *pArr;
	pArr = (TypeCellML *) argv[0];
	cuda_set_stim(pArr);
	return 1;
}


int IDL_set_DiffParm(int argc, void *argv[])
{
	TypeCellML *pCoeff;
	int *pInd;
	pInd = (int *)argv[0];
	pCoeff = (TypeCellML *)argv[1];
	cuda_set_DiffParm(pInd, pCoeff);
	return 1;
}

int IDL_calc_diffusion(int argc, void *argv[])
{
	cuda_calc_Diffusion();
	return 1;
}


int IDL_simulate_SingleStep(int argc, void *argv[])
{
	cuda_simulate_SingleStep();
	return 1;
}


int IDL_simulate_Nsteps(int argc, void *argv[])
{
	int nsteps;
	nsteps = *(int *)argv[0];
	cuda_simulate_Nsteps(nsteps);
	return 1;
}

int IDL_simulate_NstepsNoDiffusion(int argc, void *argv[])
{
	int nsteps;
	nsteps = *(int *)argv[0];
	cuda_simulate_NstepsNoDiffusion(nsteps);
	return 1;
}


