#include <stdlib.h>
#include <string.h>
#include <cuda.h>
#include <stdio.h>

#include "StcCell.h"

int cell_alloc_StcCellOnCuda(StcCell *pStc)
{	cudaError_t e;
	int res;
	int iByteSize;
	res = 0;
	if (pStc->isAllocated == 1) cell_free_StcCellOnCuda(pStc);
	// state
	iByteSize = pStc->N_CELL * pStc->N_STATE * sizeof(TypeCellML) ;
	e = cudaMalloc( (void **)&(pStc->d_State), iByteSize);
	if (e == cudaErrorMemoryAllocation) res += -1;
	if (e == cudaSuccess)  pStc->iMemFlagState = 1;

	// rate
	iByteSize = pStc->N_CELL * pStc->N_RATE * sizeof(TypeCellML) ;
	if (iByteSize > 0) {
		e = cudaMalloc( (void **)&(pStc->d_Rate), iByteSize);
		if (e ==cudaErrorMemoryAllocation) res += -2;
		if (e == cudaSuccess) pStc->iMemFlagRate = 1;
	}
	// algebraic
	
	// const
	iByteSize = pStc->N_CELL * pStc->N_CONST * sizeof(TypeCellML) ;
	e = cudaMalloc( (void **)&(pStc->d_Const), iByteSize) ;
	if (e == cudaErrorMemoryAllocation) res += -4;
	if (e == cudaSuccess) pStc->iMemFlagConst = 1;
	// cell location
	iByteSize = pStc->N_CELL * sizeof(int) ;
	e = cudaMalloc( (void **)&(pStc->d_CellLoc), iByteSize);
	if (e == cudaErrorMemoryAllocation) res += -8;
	if (e == cudaSuccess) pStc->iMemFlagCellLoc = 1;
	// temporary memory
	iByteSize = pStc->N_CELL * sizeof(TypeCellML) ;
	e = cudaMalloc( (void **)&(pStc->d_temp), iByteSize);
	if (e == cudaErrorMemoryAllocation) res += -16;
	if (e == cudaSuccess) pStc->iMemFlagTemp = 1;
	
		

	if (res == 0) pStc->isAllocated = 1;
	return res;
}


void cell_free_StcCellOnCuda(StcCell *pStc)
{
	if( pStc->iMemFlagCellLoc == 1) {
		cudaFree(pStc->d_CellLoc);
		pStc->iMemFlagCellLoc = 0;
	}
	if( pStc->iMemFlagState == 1) {
		cudaFree(pStc->d_State);
		pStc->iMemFlagState = 0;
	}
	if( pStc->iMemFlagRate== 1) {
		cudaFree(pStc->d_Rate);
		pStc->iMemFlagRate = 0;
	}
	
	if( pStc->iMemFlagConst == 1) {
		cudaFree(pStc->d_Const);
		pStc->iMemFlagConst = 0;
	}
	if (pStc->iMemFlagTemp == 1) {
		cudaFree(pStc->d_temp);
		pStc->iMemFlagTemp = 0;
	}
}




void cell_copy_VarToCuda(StcCell *pStc, TypeCellML *val, enum CellMLVarType iType)  
{
	void *d_ptr ; // TypeCellML *d_ptr;
	int iByteSize;

	switch (iType) {
		case CellLoc: 
			d_ptr = (void *)pStc->d_CellLoc;
			iByteSize = sizeof(int);
		   	break;
		case State: 
			d_ptr = (void *)pStc->d_State;
			iByteSize = pStc->N_STATE * sizeof(TypeCellML);
			break;
		case Const: 
			d_ptr = (void *)pStc->d_Const;
			iByteSize = pStc->N_CONST * sizeof(TypeCellML);
			break;
	}
	iByteSize = iByteSize * pStc->N_CELL;
	cudaMemcpy(d_ptr, val, iByteSize, cudaMemcpyHostToDevice);
}


void cell_copy_VarFromCuda(StcCell *pStc, TypeCellML *val, enum CellMLVarType iType) 
{	
	void *d_ptr ; // TypeCellML *d_ptr;
	int iByteSize;

	switch (iType) {
		case CellLoc: // no method for flag 
			d_ptr = (void *)pStc->d_CellLoc;
			iByteSize = pStc->N_CELL * sizeof(int) ;
		   	break;
		case State: 
			d_ptr = (void *)pStc->d_State;
			iByteSize = pStc->N_STATE * sizeof(TypeCellML);
			break;
		case Const: 
			d_ptr = (void *)pStc->d_Const;
			iByteSize = pStc->N_CONST * sizeof(TypeCellML);
			break;
	}
	iByteSize = iByteSize * pStc->N_CELL;
	cudaMemcpy(val, d_ptr, iByteSize, cudaMemcpyDeviceToHost);
}


			
// set values 

// setting values
__global__ void kern_set_Value(TypeCellML *d_cell, TypeCellML *d_input, int loc, int nElm, int iTotal) {
	int idx;
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		d_cell[idx * nElm + loc] = d_input[idx];
	}
}

__global__ void kern_get_Value(TypeCellML *d_cell, TypeCellML *d_temp, int loc, int nElm, int iTotal) {
	int idx;
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		d_temp[idx] = d_cell[idx * nElm + loc];
	}
}

// warining
// do not use these function for CellLoc
// use cell_copy_ ... function instead
void cell_set_CellValue(StcCell *pStc, TypeCellML *val, int loc, enum CellMLVarType iType) {
	int iBlockDim;
	TypeCellML *d_ptr, *d_temp;
	int iElm;
	d_temp = pStc->d_temp;
	// copy values to temporary memory first and copy again
	cudaMemcpy(pStc->d_temp, val, pStc->N_CELL * sizeof(TypeCellML), cudaMemcpyHostToDevice);
	iBlockDim = (pStc->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStc->N_CELL) iBlockDim ++;
	// copy to d_vtemp and copy to out array
	switch (iType) { 
			
		case Rate: 
			d_ptr = pStc->d_Rate;
			iElm = pStc->N_RATE;
			break;
		case State: 
			d_ptr = pStc->d_State;
			iElm = pStc->N_STATE;
			break;
		case Const: 
			d_ptr = pStc->d_Const;
			iElm = pStc->N_CONST; 
			break;
	}
	if (iType > 0) {
		kern_set_Value <<< iBlockDim, iCudaBlockSize >>> 
			(d_ptr, d_temp, loc, iElm, pStc->N_CELL);
	}
}


void cell_get_CellValue(StcCell *pStc, TypeCellML *val, int loc, enum CellMLVarType iType) {
	int iBlockDim;
	TypeCellML *d_ptr, *d_temp;
	int iElm;
	d_temp = pStc->d_temp;
	iBlockDim = (pStc->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStc->N_CELL) iBlockDim ++;
	// copy to d_vtemp and copy to out array
	switch (iType) { // 0: rate, 1: state, 2: constant
		
		case Rate: 
			d_ptr = pStc->d_Rate;
			iElm = pStc->N_RATE;
			break;
		case State: 
			d_ptr = pStc->d_State;
			iElm = pStc->N_STATE;
			break;
		case Const: 
			d_ptr = pStc->d_Const;
			iElm = pStc->N_CONST; 
			break;
	}
	if (iType > 0) {
		kern_get_Value <<< iBlockDim, iCudaBlockSize >>> 
			(d_ptr, d_temp, loc, iElm, pStc->N_CELL);
	}
	// copy values from temporary memory
	cudaMemcpy(val, pStc->d_temp, pStc->N_CELL * sizeof(TypeCellML), cudaMemcpyDeviceToHost);

}


// setting values loc-specific
__global__ void kern_set_Value_Loc(TypeCellML *d_cell, TypeCellML *d_input, int *d_Loc, int iVarInd, int nElm, int iTotal) {
	int idx;
	int iLocCell;
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		iLocCell = d_Loc[idx];
		d_cell[idx * nElm + iVarInd] = d_input[iLocCell];
	}
}

__global__ void kern_get_Value_Loc(TypeCellML *d_cell, TypeCellML *d_temp, int *d_Loc, int iVarInd, int nElm, int iTotal) {
	int idx;
	int iLocCell;
	idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (idx < iTotal) {
		iLocCell = d_Loc[idx];
		d_temp[iLocCell] = d_cell[idx * nElm + iVarInd];
	}
}



void cell_get_CellValueSorted(StcCell *pStc, TypeCellML *d_val, int iVarInd, enum CellMLVarType iType) {
	int iBlockDim;
	TypeCellML *d_ptr;
	int iElm;
	iBlockDim = (pStc->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStc->N_CELL) iBlockDim ++;
	// copy to d_vtemp and copy to out array
	switch (iType) { // 0: rate, 1: state, 2: constant

		case State: 
			d_ptr = pStc->d_State;
			iElm = pStc->N_STATE;
			break;
		case Const: 
			d_ptr = pStc->d_Const;
			iElm = pStc->N_CONST; 
			break;
	}
	if (iType > 0) {
		kern_get_Value_Loc <<< iBlockDim, iCudaBlockSize >>> 
			(d_ptr, d_val, pStc->d_CellLoc, iVarInd, iElm, pStc->N_CELL);
	}
	// copy values from temporary memory

}


void cell_set_CellValueSorted(StcCell *pStc, TypeCellML *d_val, int iVarInd, enum CellMLVarType iType) {
	int iBlockDim;
	TypeCellML *d_ptr;
	int iElm;
	iBlockDim = (pStc->N_CELL / iCudaBlockSize);
	if ((iBlockDim*iCudaBlockSize) < pStc->N_CELL) iBlockDim ++;
	// copy to d_vtemp and copy to out array
	switch (iType) { // 0: rate, 1: state, 2: constant

		case State: 
			d_ptr = pStc->d_State;
			iElm = pStc->N_STATE;
			break;
		case Const: 
			d_ptr = pStc->d_Const;
			iElm = pStc->N_CONST; 
			break;
	}
	kern_set_Value_Loc <<< iBlockDim, iCudaBlockSize >>> 
			(d_ptr, d_val, pStc->d_CellLoc, iVarInd, iElm, pStc->N_CELL);
	// copy values from temporary memory
	
}

void cell_get_VmSorted(StcCell *pStcCell, TypeCellML *d_out)
{
	cell_get_CellValueSorted(pStcCell, d_out, pStcCell->LOC_Vm, State);
}

void cell_set_VmDiffSorted(StcCell *pStcCell, TypeCellML *d_in)
{
	cell_set_CellValueSorted(pStcCell, d_in, pStcCell->LOC_DIFF, State);
}

